"""Constants for standard whylogs metrics.

Docs:
https://whylogs.readthedocs.io/en/latest/examples/basic/Inspecting_Profiles.html#Understanding-The-whylogs-Profile-Statistics

By default, whylogs will track the following metrics according to the column's inferred data type:

Integral:
- counts
- types
- distribution
- ints
- cardinality
- frequent_items

Fractional:
- counts
- types
- cardinality
- distribution

String:
- counts
- types
- cardinality
- frequent_items
"""

# The total number of entries in a feature.
COUNTS_N = "counts/n"
# The number of null values.
COUNTS_NULL = "counts/null"
# The number of values consisting of an integral (whole number).
TYPES_INTEGRAL = "types/integral"
# The number of values consisting of a fractional value (float).
TYPES_FRACTIONAL = "types/fractional"
# The number of values consisting of a boolean.
TYPES_BOOLEAN = "types/boolean"
# The number of values consisting of a string.
TYPES_STRING = "types/string"
# The number of values consisting of an object. If the data is not of any of the previous types, it will be assumed as
# an object.
TYPES_OBJECT = "types/object"

# Cardinality
# The estimated unique values for each feature
CARDINALITY_EST = "cardinality/est"
# Upper bound for the cardinality estimation. The actual cardinality will always be below this number.
CARDINALITY_UPPER = "cardinality/upper_1"
# Lower bound for the cardinality estimation. The actual cardinality will always be above this number.
CARDINALITY_LOWER = "cardinality/lower_1"

# The most frequent items.
FREQUENT_ITEMS = "frequent_items/frequent_strings"

# Distribution statistics are generated when a feature contains numerical data.
# The calculated mean of the feature data.
DISTRIBUTION_MEAN = "distribution/mean"
# The calculated standard deviation of the feature data.
DISTRIBUTION_STDDEV = "distribution/stddev"
# The number of rows belonging to the feature.
DISTRIBUTION_N = "distribution/n"
# The highest (max) value in the feature.
DISTRIBUTION_MAX = "distribution/max"
# The smallest (min) value in the feature.
DISTRIBUTION_MIN = "distribution/min"
# The median value of the feature data.
DISTRIBUTION_MEDIAN = "distribution/median"
# The xx-th quantile value of the data’s distribution
DISTRIBUTION_Q_XX = "distribution/q_xx"



import pytest

from ludwig.modules.reduction_modules import reduce_mode_registry
from tests.integration_tests.utils import category_feature, generate_data, run_experiment, sequence_feature


@pytest.mark.parametrize("reduce_output", reduce_mode_registry)
def test_reduction(reduce_output, csv_filename):
    input_features = [sequence_feature(reduce_output=reduce_output)]

    output_features = [category_feature(output_feature=True)]

    rel_path = generate_data(input_features, output_features, csv_filename)
    run_experiment(input_features, output_features, dataset=rel_path)

    del input_features
    del output_features



import pytest
import torch

from ludwig.encoders.generic_encoders import DenseEncoder, PassthroughEncoder


@pytest.mark.parametrize("input_size", [1, 2, 10])
@pytest.mark.parametrize("categorical", [True, False])
def test_generic_passthrough_encoder(input_size: int, categorical: bool):
    passthrough_encoder = PassthroughEncoder(input_size)
    # Passthrough encoder allows categorical input feature (int), dense encoder's input must be float.
    if categorical:
        inputs = torch.randint(10, (10, input_size))
    else:
        inputs = torch.rand((10, input_size))
    outputs = passthrough_encoder(inputs)
    # Ensures output shape matches encoder expected output shape.
    assert outputs["encoder_output"].shape[1:] == passthrough_encoder.output_shape


@pytest.mark.parametrize("input_size", [1, 2, 10])
@pytest.mark.parametrize("num_layers", [1, 3, 6])
@pytest.mark.parametrize("output_size", [1, 2, 10, 256])
def test_generic_dense_encoder(input_size: int, num_layers: int, output_size: int):
    dense_encoder = DenseEncoder(input_size, num_layers=num_layers, output_size=output_size)
    inputs = torch.rand((10, input_size))
    outputs = dense_encoder(inputs)
    # Ensures output shape matches encoder expected output shape.
    assert outputs["encoder_output"].shape[1:] == dense_encoder.output_shape



import numpy as np
import pandas as pd
import pytest

from ludwig.schema.features.preprocessing.text import TextPreprocessingConfig
from ludwig.utils import strings_utils


def test_is_number():
    assert strings_utils.is_number("1.1")
    assert strings_utils.is_number("1.000001")
    assert strings_utils.is_number("1000001")
    assert strings_utils.is_number("Nan")
    assert strings_utils.is_number("NaN")
    assert strings_utils.is_number(1)
    assert strings_utils.is_number(1.1)
    assert not strings_utils.is_number("NaNaaa")


def test_are_sequential_integers():
    assert strings_utils.are_sequential_integers(["1.0", "2", "3"])
    assert strings_utils.are_sequential_integers(["1", "2", "3"])
    assert not strings_utils.are_sequential_integers(["1", "2", "4"])
    assert not strings_utils.are_sequential_integers(["1.1", "2", "3"])
    assert not strings_utils.are_sequential_integers(["a", "2", "3"])


def test_str_to_bool():
    # Global bool mappings are used.
    assert strings_utils.str2bool("True")
    assert strings_utils.str2bool(True)
    assert strings_utils.str2bool("true")
    assert not strings_utils.str2bool("0")

    # Error raised if non-mapped value is encountered and no fallback is specified.
    with pytest.raises(Exception):
        strings_utils.str2bool("bot")

    # Fallback label is used.
    assert strings_utils.str2bool("bot", fallback_true_label="bot")
    assert not strings_utils.str2bool("human", fallback_true_label="bot")
    assert strings_utils.str2bool("human", fallback_true_label="human")
    assert not strings_utils.str2bool("human", fallback_true_label="Human")

    # Fallback label is used, strictly as a fallback.
    assert strings_utils.str2bool("True", fallback_true_label="False")


def test_are_conventional_bools():
    assert strings_utils.are_conventional_bools(["True", "False"])
    assert strings_utils.are_conventional_bools([True, False])
    assert strings_utils.are_conventional_bools(["True", False, True])
    assert strings_utils.are_conventional_bools(["T", "F"])
    assert strings_utils.are_conventional_bools(["t", "f"])
    assert not strings_utils.are_conventional_bools(["True", "Fails"])
    assert strings_utils.are_conventional_bools(["0", "1"])
    assert not strings_utils.are_conventional_bools(["0", "2"])
    assert strings_utils.are_conventional_bools(["1.0", "0.0"])
    assert not strings_utils.are_conventional_bools(["high", "low"])
    assert not strings_utils.are_conventional_bools(["human", "bot"])


def test_create_vocabulary_chars():
    data = pd.DataFrame(["Hello, I'm a single sentence!", "And another sentence", "And the very very last one"])
    column = data[0]
    preprocessing_parameters = TextPreprocessingConfig().to_dict()

    vocabulary_output = strings_utils.create_vocabulary(
        column,
        tokenizer_type="characters",
        num_most_frequent=preprocessing_parameters["most_common"],
        lowercase=preprocessing_parameters["lowercase"],
        unknown_symbol=preprocessing_parameters["unknown_symbol"],
        padding_symbol=preprocessing_parameters["padding_symbol"],
        pretrained_model_name_or_path=preprocessing_parameters["pretrained_model_name_or_path"],
    )

    assert len(vocabulary_output[0]) == 24
    assert vocabulary_output[0][strings_utils.SpecialSymbol.START.value] == strings_utils.START_SYMBOL
    assert vocabulary_output[0][strings_utils.SpecialSymbol.STOP.value] == strings_utils.STOP_SYMBOL
    assert vocabulary_output[0][strings_utils.SpecialSymbol.PADDING.value] == strings_utils.PADDING_SYMBOL
    assert vocabulary_output[0][strings_utils.SpecialSymbol.UNKNOWN.value] == strings_utils.UNKNOWN_SYMBOL


def test_create_vocabulary_word():
    data = pd.DataFrame(["Hello, I'm a single sentence!", "And another sentence", "And the very very last one"])
    column = data[0]
    preprocessing_parameters = TextPreprocessingConfig().to_dict()

    vocabulary_output = strings_utils.create_vocabulary(
        column,
        tokenizer_type=preprocessing_parameters["tokenizer"],
        num_most_frequent=preprocessing_parameters["most_common"],
        lowercase=preprocessing_parameters["lowercase"],
        vocab_file=preprocessing_parameters["vocab_file"],
        unknown_symbol=preprocessing_parameters["unknown_symbol"],
        padding_symbol=preprocessing_parameters["padding_symbol"],
        pretrained_model_name_or_path=preprocessing_parameters["pretrained_model_name_or_path"],
    )

    assert len(vocabulary_output[0]) == 19
    assert vocabulary_output[0][strings_utils.SpecialSymbol.UNKNOWN.value] == strings_utils.UNKNOWN_SYMBOL
    assert vocabulary_output[0][strings_utils.SpecialSymbol.STOP.value] == strings_utils.STOP_SYMBOL
    assert vocabulary_output[0][strings_utils.SpecialSymbol.PADDING.value] == strings_utils.PADDING_SYMBOL
    assert vocabulary_output[0][strings_utils.SpecialSymbol.UNKNOWN.value] == strings_utils.UNKNOWN_SYMBOL


def test_create_vocabulary_no_special_symbols():
    data = pd.DataFrame(["Hello, I'm a single sentence!", "And another sentence", "And the very very last one"])
    column = data[0]
    preprocessing_parameters = TextPreprocessingConfig().to_dict()

    vocabulary_output = strings_utils.create_vocabulary(
        column,
        tokenizer_type=preprocessing_parameters["tokenizer"],
        num_most_frequent=preprocessing_parameters["most_common"],
        lowercase=preprocessing_parameters["lowercase"],
        vocab_file=preprocessing_parameters["vocab_file"],
        unknown_symbol=preprocessing_parameters["unknown_symbol"],
        padding_symbol=preprocessing_parameters["padding_symbol"],
        pretrained_model_name_or_path=preprocessing_parameters["pretrained_model_name_or_path"],
        add_special_symbols=False,
    )

    assert len(vocabulary_output[0]) == 16
    assert vocabulary_output[0][strings_utils.SpecialSymbol.UNKNOWN.value] == strings_utils.UNKNOWN_SYMBOL


def test_create_vocabulary_from_hf():
    data = pd.DataFrame(["Hello, I'm a single sentence!", "And another sentence", "And the very very last one"])
    column = data[0]
    preprocessing_parameters = TextPreprocessingConfig().to_dict()

    vocabulary_output = strings_utils.create_vocabulary(
        column,
        tokenizer_type="hf_tokenizer",
        num_most_frequent=preprocessing_parameters["most_common"],
        lowercase=preprocessing_parameters["lowercase"],
        unknown_symbol=preprocessing_parameters["unknown_symbol"],
        padding_symbol=preprocessing_parameters["padding_symbol"],
        pretrained_model_name_or_path="albert-base-v2",
    )

    assert len(vocabulary_output[0]) == 30000


def test_create_vocabulary_single_token():
    data = pd.DataFrame(["dog", "cat", "bird", "dog", "cat", "super cat"])
    column = data[0]

    vocab, str2idx, str2freq = strings_utils.create_vocabulary_single_token(
        column,
        num_most_frequent=10000,
    )

    assert set(vocab) == {"dog", "cat", "bird", "super cat"}
    assert str2freq == {"dog": 2, "cat": 2, "bird": 1, "super cat": 1}
    assert strings_utils.UNKNOWN_SYMBOL not in str2idx


def test_create_vocabulary_single_token_small_most_frequent():
    data = pd.DataFrame(["dog", "cat", "bird", "dog", "cat", "super cat"])
    column = data[0]

    vocab, str2idx, str2freq = strings_utils.create_vocabulary_single_token(column, num_most_frequent=2)

    assert set(vocab) == {"dog", "cat", strings_utils.UNKNOWN_SYMBOL}
    assert str2idx[strings_utils.UNKNOWN_SYMBOL] == 0
    assert str2freq == {"dog": 2, "cat": 2, strings_utils.UNKNOWN_SYMBOL: 0}


def test_build_sequence_matrix():
    inverse_vocabulary = {
        "<EOS>": 0,
        "<SOS>": 1,
        "<PAD>": 2,
        "<UNK>": 3,
        "a": 4,
        "b": 5,
        "c": 6,
    }
    sequences = pd.core.series.Series(["a b c", "c b a"])
    sequence_matrix = strings_utils.build_sequence_matrix(
        sequences, inverse_vocabulary, tokenizer_type="space", length_limit=10
    )
    assert not (
        sequence_matrix.tolist() - np.array([[1, 4, 5, 6, 0, 2, 2, 2, 2, 2], [1, 6, 5, 4, 0, 2, 2, 2, 2, 2]])
    ).any()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ludwig/profiling/proto/dataset_profile.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ludwig.profiling.proto import whylogs_messages_pb2 as ludwig_dot_profiling_dot_proto_dot_whylogs__messages__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='ludwig/profiling/proto/dataset_profile.proto',
  package='dataset_profile',
  syntax='proto3',
  serialized_options=b'Z+github.com/ludwig-ai/ludwig/dataset_profile',
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n,ludwig/profiling/proto/dataset_profile.proto\x12\x0f\x64\x61taset_profile\x1a-ludwig/profiling/proto/whylogs_messages.proto\"\xf6\x01\n\x0e\x44\x61tasetProfile\x12\x11\n\ttimestamp\x18\x01 \x01(\x03\x12\x14\n\x0cnum_examples\x18\x02 \x01(\x03\x12\x12\n\nsize_bytes\x18\x03 \x01(\x03\x12N\n\x10\x66\x65\x61ture_profiles\x18\x14 \x03(\x0b\x32\x34.dataset_profile.DatasetProfile.FeatureProfilesEntry\x1aW\n\x14\x46\x65\x61tureProfilesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12.\n\x05value\x18\x02 \x01(\x0b\x32\x1f.dataset_profile.FeatureProfile:\x02\x38\x01\"I\n\x0e\x46\x65\x61tureProfile\x12\x37\n\x0fwhylogs_metrics\x18\x01 \x01(\x0b\x32\x1e.dataset_profile.ColumnMessageB-Z+github.com/ludwig-ai/ludwig/dataset_profileb\x06proto3'
  ,
  dependencies=[ludwig_dot_profiling_dot_proto_dot_whylogs__messages__pb2.DESCRIPTOR,])




_DATASETPROFILE_FEATUREPROFILESENTRY = _descriptor.Descriptor(
  name='FeatureProfilesEntry',
  full_name='dataset_profile.DatasetProfile.FeatureProfilesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='dataset_profile.DatasetProfile.FeatureProfilesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='value', full_name='dataset_profile.DatasetProfile.FeatureProfilesEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=b'8\001',
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=272,
  serialized_end=359,
)

_DATASETPROFILE = _descriptor.Descriptor(
  name='DatasetProfile',
  full_name='dataset_profile.DatasetProfile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='dataset_profile.DatasetProfile.timestamp', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='num_examples', full_name='dataset_profile.DatasetProfile.num_examples', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='size_bytes', full_name='dataset_profile.DatasetProfile.size_bytes', index=2,
      number=3, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='feature_profiles', full_name='dataset_profile.DatasetProfile.feature_profiles', index=3,
      number=20, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[_DATASETPROFILE_FEATUREPROFILESENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=113,
  serialized_end=359,
)


_FEATUREPROFILE = _descriptor.Descriptor(
  name='FeatureProfile',
  full_name='dataset_profile.FeatureProfile',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='whylogs_metrics', full_name='dataset_profile.FeatureProfile.whylogs_metrics', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=361,
  serialized_end=434,
)

_DATASETPROFILE_FEATUREPROFILESENTRY.fields_by_name['value'].message_type = _FEATUREPROFILE
_DATASETPROFILE_FEATUREPROFILESENTRY.containing_type = _DATASETPROFILE
_DATASETPROFILE.fields_by_name['feature_profiles'].message_type = _DATASETPROFILE_FEATUREPROFILESENTRY
_FEATUREPROFILE.fields_by_name['whylogs_metrics'].message_type = ludwig_dot_profiling_dot_proto_dot_whylogs__messages__pb2._COLUMNMESSAGE
DESCRIPTOR.message_types_by_name['DatasetProfile'] = _DATASETPROFILE
DESCRIPTOR.message_types_by_name['FeatureProfile'] = _FEATUREPROFILE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

DatasetProfile = _reflection.GeneratedProtocolMessageType('DatasetProfile', (_message.Message,), {

  'FeatureProfilesEntry' : _reflection.GeneratedProtocolMessageType('FeatureProfilesEntry', (_message.Message,), {
    'DESCRIPTOR' : _DATASETPROFILE_FEATUREPROFILESENTRY,
    '__module__' : 'ludwig.profiling.proto.dataset_profile_pb2'
    # @@protoc_insertion_point(class_scope:dataset_profile.DatasetProfile.FeatureProfilesEntry)
    })
  ,
  'DESCRIPTOR' : _DATASETPROFILE,
  '__module__' : 'ludwig.profiling.proto.dataset_profile_pb2'
  # @@protoc_insertion_point(class_scope:dataset_profile.DatasetProfile)
  })
_sym_db.RegisterMessage(DatasetProfile)
_sym_db.RegisterMessage(DatasetProfile.FeatureProfilesEntry)

FeatureProfile = _reflection.GeneratedProtocolMessageType('FeatureProfile', (_message.Message,), {
  'DESCRIPTOR' : _FEATUREPROFILE,
  '__module__' : 'ludwig.profiling.proto.dataset_profile_pb2'
  # @@protoc_insertion_point(class_scope:dataset_profile.FeatureProfile)
  })
_sym_db.RegisterMessage(FeatureProfile)


DESCRIPTOR._options = None
_DATASETPROFILE_FEATUREPROFILESENTRY._options = None
# @@protoc_insertion_point(module_scope)



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import math
from typing import List

import numpy as np


def softmax(x, temperature=1.0):
    e_x = np.exp((x - np.max(x)) / temperature)
    return e_x / e_x.sum()


def int_type(number):
    if number <= np.iinfo(np.int8).max:
        return np.int8
    elif number <= np.iinfo(np.int16).max:
        return np.int16
    elif number <= np.iinfo(np.int32).max:
        return np.int32
    else:  # if number <= np.iinfo(np.int64).max:
        return np.int64


def convert_size(size_bytes):
    if size_bytes == 0:
        return "0B"
    size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
    i = int(math.floor(math.log(size_bytes, 1024)))
    p = math.pow(1024, i)
    s = round(size_bytes / p, 2)
    return f"{s} {size_name[i]}"


def round2precision(val, precision: int = 0, which: str = ""):
    assert precision >= 0
    val *= 10**precision
    round_callback = round
    if which.lower() == "up":
        round_callback = math.ceil
    if which.lower() == "down":
        round_callback = math.floor
    return "{1:.{0}f}".format(precision, round_callback(val) / 10**precision)


def cumsum(x: List[int]) -> List[int]:
    results = []
    j = 0
    for i in range(0, len(x)):
        j += x[i]
        results.append(j)
    return results



from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import pandas as pd
import ray
from torch.autograd import Variable
from tqdm import tqdm

from ludwig.api import LudwigModel
from ludwig.api_annotations import PublicAPI
from ludwig.explain.captum import (
    ExplanationRunConfig,
    get_baseline,
    get_input_tensors,
    get_total_attribution,
    IntegratedGradientsExplainer,
    retry_on_cuda_oom,
)
from ludwig.explain.explanation import ExplanationsResult
from ludwig.features.feature_utils import LudwigFeatureDict
from ludwig.utils.torch_utils import get_torch_device


@PublicAPI(stability="experimental")
class RayIntegratedGradientsExplainer(IntegratedGradientsExplainer):
    def __init__(self, *args, resources_per_task: Dict[str, Any] = None, num_workers: int = 1, **kwargs):
        super().__init__(*args, **kwargs)
        self.resources_per_task = resources_per_task or {}
        self.num_workers = num_workers

    def explain(self) -> ExplanationsResult:
        """Explain the model's predictions using Integrated Gradients.

        # Return

        :return: ExplanationsResult containing the explanations.
            `global_explanations`: (Explanation) Aggregate explanation for the entire input data.

            `row_explanations`: (List[Explanation]) A list of explanations, one for each row in the input data. Each
            explanation contains the integrated gradients for each label in the target feature's vocab with respect to
            each input feature.

            `expected_values`: (List[float]) of length [output feature cardinality] Average convergence delta for each
            label in the target feature's vocab.
        """
        self.model.model.cpu()
        input_features: LudwigFeatureDict = self.model.model.input_features
        model_ref = ray.put(self.model)
        run_config = ExplanationRunConfig(batch_size=self.model.config_obj.trainer.batch_size)

        # Convert input data into embedding tensors from the output of the model encoders.
        inputs_encoded_ref = get_input_tensors_task.options(**self.resources_per_task).remote(
            model_ref, ray.put(self.inputs_df), run_config
        )
        sample_encoded_ref = get_input_tensors_task.options(**self.resources_per_task).remote(
            model_ref, ray.put(self.sample_df), run_config
        )

        inputs_encoded, run_config = ray.get(inputs_encoded_ref)
        sample_encoded, run_config = ray.get(sample_encoded_ref)
        baseline = get_baseline(self.model, sample_encoded)

        inputs_encoded_ref = ray.put(inputs_encoded)
        baseline_ref = ray.put(baseline)

        if self.is_category_target:
            # Evenly divide the list of labels among the desired number of workers (Ray tasks).
            # For example, 4 GPUs -> 4 workers. We do this instead of creating nlabels tasks because
            # there is significant overhead to spawning a Ray task.
            target_splits = split_list(list(range(self.vocab_size)), self.num_workers)
        else:
            # No target index to compare against exists for number features.
            # For binary targets, we only need to compute attribution for the positive class (see below).
            # May need to revisit in the future for additional feature types.
            target_splits = [[None]]

        # Compute attribution for each possible output feature label separately.
        attrs_refs = []
        for target_indices in target_splits:
            attrs_ref = get_total_attribution_task.options(**self.resources_per_task).remote(
                model_ref,
                self.target_feature_name,
                target_indices,
                inputs_encoded_ref,
                baseline_ref,
                len(self.inputs_df),
                run_config,
            )
            attrs_refs.append(attrs_ref)

        # Await the completion of our Ray tasks, then merge the results.
        expected_values = []
        for attrs_ref in tqdm(attrs_refs, desc="Explain"):
            attrs = ray.get(attrs_ref)
            for total_attribution, feat_to_token_attributions, total_attribution_global in attrs:
                # Aggregate token attributions
                feat_to_token_attributions_global = {}
                for feat_name, token_attributions in feat_to_token_attributions.items():
                    token_attributions_global = defaultdict(float)
                    # sum attributions for each token
                    for token, token_attribution in (ta for tas in token_attributions for ta in tas):
                        token_attributions_global[token] += token_attribution
                    # divide by number of samples to get average attribution per token
                    token_attributions_global = {
                        token: token_attribution / max(0, len(token_attributions))
                        for token, token_attribution in token_attributions_global.items()
                    }
                    # convert to list of tuples and sort by attribution
                    token_attributions_global = sorted(
                        token_attributions_global.items(), key=lambda x: x[1], reverse=True
                    )
                    # keep only top 100 tokens
                    token_attributions_global = token_attributions_global[:100]
                    feat_to_token_attributions_global[feat_name] = token_attributions_global

                self.global_explanation.add(
                    input_features.keys(), total_attribution_global, feat_to_token_attributions_global
                )

                for i, (feature_attributions, explanation) in enumerate(zip(total_attribution, self.row_explanations)):
                    # Add the feature attributions to the explanation object for this row.
                    explanation.add(
                        input_features.keys(),
                        feature_attributions,
                        {k: v[i] for k, v in feat_to_token_attributions.items()},
                    )

                # TODO(travis): for force plots, need something similar to SHAP E[X]
                expected_values.append(0.0)

        # For binary targets, add an extra attribution for the negative class (false).
        if self.is_binary_target:
            le_true = self.global_explanation.label_explanations[0]
            negated_attributions = le_true.to_array() * -1
            negated_token_attributions = {
                fa.feature_name: [(t, -a) for t, a in fa.token_attributions]
                for fa in le_true.feature_attributions
                if fa.token_attributions is not None
            }
            # Prepend the negative class to the list of label explanations.
            self.global_explanation.add(
                input_features.keys(), negated_attributions, negated_token_attributions, prepend=True
            )

            for explanation in self.row_explanations:
                le_true = explanation.label_explanations[0]
                negated_attributions = le_true.to_array() * -1
                negated_token_attributions = {
                    fa.feature_name: [(t, -a) for t, a in fa.token_attributions]
                    for fa in le_true.feature_attributions
                    if fa.token_attributions is not None
                }
                # Prepend the negative class to the list of label explanations.
                explanation.add(input_features.keys(), negated_attributions, negated_token_attributions, prepend=True)

            # TODO(travis): for force plots, need something similar to SHAP E[X]
            expected_values.append(0.0)

        return ExplanationsResult(self.global_explanation, self.row_explanations, expected_values)


@ray.remote(max_calls=1)
def get_input_tensors_task(
    model: LudwigModel, df: pd.DataFrame, run_config: ExplanationRunConfig
) -> Tuple[List[Variable], ExplanationRunConfig]:
    model.model.unskip()
    model.model.to(get_torch_device())
    try:
        get_total_attribution_with_retry = retry_on_cuda_oom(run_config)(get_input_tensors)
        return get_total_attribution_with_retry(model, df, run_config), run_config
    finally:
        model.model.cpu()


@ray.remote(max_calls=1)
def get_total_attribution_task(
    model: LudwigModel,
    target_feature_name: str,
    target_indices: List[Optional[int]],
    inputs_encoded: List[Variable],
    baseline: List[Variable],
    nsamples: int,
    run_config: ExplanationRunConfig,
) -> List[np.array]:
    model.model.unskip()
    model.model.to(get_torch_device())
    try:
        get_total_attribution_with_retry = retry_on_cuda_oom(run_config)(get_total_attribution)
        return [
            get_total_attribution_with_retry(
                model=model,
                target_feature_name=target_feature_name,
                target_idx=target_idx,
                feature_inputs=inputs_encoded,
                baseline=baseline,
                nsamples=nsamples,
                run_config=run_config,
            )
            for target_idx in tqdm(target_indices, desc="Explain")
        ]
    finally:
        model.model.cpu()


def split_list(v, n):
    """Splits a list into n roughly equal sub-lists.

    Source: https://stackoverflow.com/a/2135920
    """
    k, m = divmod(len(v), n)
    return (v[i * k + min(i, m) : (i + 1) * k + min(i + 1, m)] for i in range(n))



import os
from dataclasses import dataclass
from typing import Any, Dict

from ludwig.types import ModelConfigDict, TrainingSetMetadataDict
from ludwig.utils.data_utils import load_json, load_yaml


@dataclass
class BenchmarkingResult:
    # The Ludwig benchmarking config.
    benchmarking_config: Dict[str, Any]

    # The config for one experiment.
    experiment_config: Dict[str, Any]

    # The Ludwig config used to run the experiment.
    ludwig_config: ModelConfigDict

    # The python script that is used to process the config before being used.
    process_config_file: str

    # Loaded `description.json` file.
    description: Dict[str, Any]

    # Loaded `test_statistics.json` file.
    test_statistics: Dict[str, Any]

    # Loaded `training_statistics.json` file.
    training_statistics: Dict[str, Any]

    # Loaded `model_hyperparameters.json` file.
    model_hyperparameters: Dict[str, Any]

    # Loaded `training_progress.json` file.
    training_progress: Dict[str, Any]

    # Loaded `training_set_metadata.json` file.
    training_set_metadata: TrainingSetMetadataDict


def build_benchmarking_result(benchmarking_config: dict, experiment_idx: int):
    experiment_config = benchmarking_config["experiments"][experiment_idx]
    process_config_file = ""
    if experiment_config["process_config_file_path"]:
        with open(experiment_config["process_config_file_path"]) as f:
            process_config_file = "".join(f.readlines())
    experiment_run_path = os.path.join(experiment_config["experiment_name"], "experiment_run")

    return BenchmarkingResult(
        benchmarking_config=benchmarking_config,
        experiment_config=experiment_config,
        ludwig_config=load_yaml(experiment_config["config_path"]),
        process_config_file=process_config_file,
        description=load_json(os.path.join(experiment_run_path, "description.json")),
        test_statistics=load_json(os.path.join(experiment_run_path, "test_statistics.json")),
        training_statistics=load_json(os.path.join(experiment_run_path, "training_statistics.json")),
        model_hyperparameters=load_json(os.path.join(experiment_run_path, "model", "model_hyperparameters.json")),
        training_progress=load_json(os.path.join(experiment_run_path, "model", "training_progress.json")),
        training_set_metadata=load_json(os.path.join(experiment_run_path, "model", "training_set_metadata.json")),
    )



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from dataclasses import asdict
from typing import Optional

import torch

from ludwig.distributed.base import DistributedStrategy
from ludwig.schema.optimizers import BaseOptimizerConfig, GradientClippingConfig, optimizer_registry, SGDOptimizerConfig
from ludwig.utils.misc_utils import get_from_registry


def create_clipper(gradient_clipping_config: Optional[GradientClippingConfig]):
    """Utility function that will convert a None-type gradient clipping config to the correct form."""
    if isinstance(gradient_clipping_config, GradientClippingConfig):
        return gradient_clipping_config
    # Return default config if provided value is None:
    return GradientClippingConfig()


def create_optimizer(
    model,
    learning_rate,
    distributed: DistributedStrategy,
    optimizer_config: BaseOptimizerConfig = SGDOptimizerConfig(),
):
    """Returns a ready-to-use torch optimizer instance based on the given optimizer config.

    :param model: Underlying Ludwig model
    :param learning_rate: Initial learning rate for the optimizer
    :param optimizer_config: Instance of `ludwig.modules.optimization_modules.BaseOptimizerConfig` (default:
           `ludwig.modules.optimization_modules.SGDOptimizerConfig()`).
    :param horovod: Horovod parameters (default: None).
    :return: Initialized instance of a torch optimizer.
    """
    # Get the corresponding torch optimizer class for the given config:
    optimizer_cls = get_from_registry(optimizer_config.type.lower(), optimizer_registry)[0]

    # Create a dict of parameters to be passed to torch (i.e. everything except `type`):
    cls_kwargs = {field: value for field, value in asdict(optimizer_config).items() if field != "type"}
    cls_kwargs["lr"] = learning_rate

    # Instantiate the optimizer:
    torch_optimizer: torch.optim.Optimizer = optimizer_cls(params=model.parameters(), **cls_kwargs)
    torch_optimizer = distributed.wrap_optimizer(torch_optimizer, model)
    return torch_optimizer



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import os

import pandas as pd
import torch

from ludwig.api import LudwigModel
from ludwig.constants import BATCH_SIZE, TRAINER
from ludwig.utils.inference_utils import to_inference_module_input_from_dataframe
from ludwig.utils.triton_utils import export_triton, get_inference_modules, POSTPROCESSOR, PREDICTOR, PREPROCESSOR
from tests.integration_tests.utils import (
    binary_feature,
    category_feature,
    generate_data,
    LocalTestBackend,
    number_feature,
    sequence_feature,
    set_feature,
    text_feature,
    vector_feature,
)


def test_triton_torchscript(csv_filename, tmpdir):
    # Configure features to be tested:
    input_features = [
        binary_feature(),
        number_feature(),
        category_feature(encoder={"vocab_size": 3}),
        # TODO: future support
        # sequence_feature(encoder={"vocab_size": 3}),
        # text_feature(encoder={"vocab_size": 3}),
        # vector_feature(),
        # timeseries_feature(),
        # date_feature(),
        # h3_feature(),
        # set_feature(encoder={"vocab_size": 3}),
        # bag_feature(encoder={"vocab_size": 3}),
        # image_feature(image_dest_folder),
        # audio_feature(audio_dest_folder),
    ]
    output_features = [
        binary_feature(),
        number_feature(),
        category_feature(decoder={"vocab_size": 3}),
        sequence_feature(decoder={"vocab_size": 3}),
        text_feature(decoder={"vocab_size": 3}),
        set_feature(decoder={"vocab_size": 3}),
        vector_feature(),
    ]
    backend = LocalTestBackend()
    config = {
        "input_features": input_features,
        "output_features": output_features,
        TRAINER: {"epochs": 1, BATCH_SIZE: 128},
    }

    # Generate training data
    training_data_csv_path = generate_data(input_features, output_features, csv_filename)

    # Train Ludwig (Pythonic) model:
    ludwig_model = LudwigModel(config, backend=backend)
    ludwig_model.train(
        dataset=training_data_csv_path,
        skip_save_training_description=True,
        skip_save_training_statistics=True,
        skip_save_model=True,
        skip_save_progress=True,
        skip_save_log=True,
        skip_save_processed_input=True,
    )

    # Create graph inference model (Torchscript) from trained Ludwig model.
    triton_path = os.path.join(tmpdir, "triton")
    model_name = "test_triton"
    model_version = "1"
    df = pd.read_csv(training_data_csv_path)
    triton_artifacts = export_triton(
        model=ludwig_model, data_example=df, model_name=model_name, output_path=triton_path, model_version=model_version
    )

    # Validate that artifact paths exist.
    assert os.path.isdir(triton_path)
    assert all(os.path.exists(artifact.path) for artifact in triton_artifacts)

    # Load TorchScript models exported for Triton.
    triton_preprocessor = triton_predictor = triton_postprocessor = None
    for artifact in triton_artifacts:
        if artifact.model_name.endswith(PREPROCESSOR) and artifact.content_type == "application/octet-stream":
            triton_preprocessor = torch.jit.load(artifact.path)
        if artifact.model_name.endswith(PREDICTOR) and artifact.content_type == "application/octet-stream":
            triton_predictor = torch.jit.load(artifact.path)
        if artifact.model_name.endswith(POSTPROCESSOR) and artifact.content_type == "application/octet-stream":
            triton_postprocessor = torch.jit.load(artifact.path)

    assert triton_preprocessor is not None
    assert triton_predictor is not None
    assert triton_postprocessor is not None

    # Forward data through models.
    data_to_predict = to_inference_module_input_from_dataframe(df, ludwig_model.config, load_paths=True, device="cpu")
    triton_preprocessor_output = triton_preprocessor(*data_to_predict.values())
    triton_predictor_output = triton_predictor(*triton_preprocessor_output)
    triton_postprocessor_output = triton_postprocessor(*triton_predictor_output)

    # Get TorchScript inference modules and forward data.
    inference_modules = get_inference_modules(ludwig_model, "cpu")
    preprocessor_output = inference_modules[0](data_to_predict)
    predictor_output = inference_modules[1](preprocessor_output)
    postprocessor_output = inference_modules[2](predictor_output)

    assert len(postprocessor_output) == len(
        triton_postprocessor_output
    ), "Number of output mismatch after postprocessor step"

    for i, (_, out_value) in enumerate(postprocessor_output.items()):
        both_list = isinstance(out_value, list) and isinstance(triton_postprocessor_output[i], list)
        both_tensor = isinstance(out_value, torch.Tensor) and isinstance(triton_postprocessor_output[i], torch.Tensor)
        assert both_list or both_tensor, "Type mismatch in PREDICTIONS, PROBABILITIES, LOGITS output"

        if isinstance(out_value, list) and len(out_value) > 0 and isinstance(out_value[0], str):
            assert out_value == triton_postprocessor_output[i], "Category feature outputs failure."
        elif isinstance(out_value, list) and len(out_value) > 0 and isinstance(out_value[0], torch.Tensor):
            assert len(out_value) == len(triton_postprocessor_output[i]), "Set feature outputs failure."
            assert all(
                torch.allclose(inf, trit) for inf, trit in zip(out_value, triton_postprocessor_output[i])
            ), "Set feature outputs failure."
        elif isinstance(out_value, list) and len(out_value) > 0 and isinstance(out_value[0], list):
            assert len(out_value) == len(
                triton_postprocessor_output[i]
            ), "Sequence (including text, etc.) feature outputs failure."
            assert all(
                inf == trit for inf, trit in zip(out_value, triton_postprocessor_output[i])
            ), "Sequence (including text, etc.) feature outputs failure."
        elif isinstance(out_value, torch.Tensor):
            assert torch.allclose(out_value, triton_postprocessor_output[i])
        else:
            raise ValueError("Value should be either List[str] or torch.Tensor.")



import copy
import logging
import os
from pprint import pformat
from typing import List, Optional, Union

import pandas as pd
import yaml
from tabulate import tabulate

from ludwig.api import LudwigModel
from ludwig.backend import Backend, initialize_backend, LocalBackend
from ludwig.callbacks import Callback
from ludwig.constants import (
    AUTO,
    COMBINED,
    EXECUTOR,
    GOAL,
    HYPEROPT,
    LOSS,
    MAX_CONCURRENT_TRIALS,
    METRIC,
    NAME,
    OUTPUT_FEATURES,
    PARAMETERS,
    PREPROCESSING,
    SEARCH_ALG,
    SPLIT,
    TEST,
    TRAINING,
    TYPE,
    VALIDATION,
)
from ludwig.data.split import get_splitter
from ludwig.hyperopt.results import HyperoptResults
from ludwig.hyperopt.utils import (
    log_warning_if_all_grid_type_parameters,
    print_hyperopt_results,
    save_hyperopt_stats,
    should_tune_preprocessing,
    update_hyperopt_params_with_defaults,
)
from ludwig.schema.model_config import ModelConfig
from ludwig.utils.backward_compatibility import upgrade_config_dict_to_latest_version
from ludwig.utils.dataset_utils import generate_dataset_statistics
from ludwig.utils.defaults import default_random_seed
from ludwig.utils.fs_utils import makedirs, open_file

try:
    from ray.tune import Callback as TuneCallback

    from ludwig.backend.ray import RayBackend
except ImportError:
    TuneCallback = object

    class RayBackend:
        pass


logger = logging.getLogger(__name__)


def hyperopt(
    config: Union[str, dict],
    dataset: Union[str, dict, pd.DataFrame] = None,
    training_set: Union[str, dict, pd.DataFrame] = None,
    validation_set: Union[str, dict, pd.DataFrame] = None,
    test_set: Union[str, dict, pd.DataFrame] = None,
    training_set_metadata: Union[str, dict] = None,
    data_format: str = None,
    experiment_name: str = "hyperopt",
    model_name: str = "run",
    resume: Optional[bool] = None,
    skip_save_training_description: bool = False,
    skip_save_training_statistics: bool = False,
    skip_save_model: bool = False,
    skip_save_progress: bool = False,
    skip_save_log: bool = False,
    skip_save_processed_input: bool = True,
    skip_save_unprocessed_output: bool = False,
    skip_save_predictions: bool = False,
    skip_save_eval_stats: bool = False,
    skip_save_hyperopt_statistics: bool = False,
    output_directory: str = "results",
    gpus: Union[str, int, List[int]] = None,
    gpu_memory_limit: Optional[float] = None,
    allow_parallel_threads: bool = True,
    callbacks: List[Callback] = None,
    tune_callbacks: List[TuneCallback] = None,
    backend: Union[Backend, str] = None,
    random_seed: int = default_random_seed,
    hyperopt_log_verbosity: int = 3,
    **kwargs,
) -> HyperoptResults:
    """This method performs an hyperparameter optimization.

    # Inputs

    :param config: (Union[str, dict]) config which defines
        the different parameters of the model, features, preprocessing and
        training.  If `str`, filepath to yaml configuration file.
    :param dataset: (Union[str, dict, pandas.DataFrame], default: `None`)
        source containing the entire dataset to be used in the experiment.
        If it has a split column, it will be used for splitting (0 for train,
        1 for validation, 2 for test), otherwise the dataset will be
        randomly split.
    :param training_set: (Union[str, dict, pandas.DataFrame], default: `None`)
        source containing training data.
    :param validation_set: (Union[str, dict, pandas.DataFrame], default: `None`)
        source containing validation data.
    :param test_set: (Union[str, dict, pandas.DataFrame], default: `None`)
        source containing test data.
    :param training_set_metadata: (Union[str, dict], default: `None`)
        metadata JSON file or loaded metadata.  Intermediate preprocessed
        structure containing the mappings of the input
        dataset created the first time an input file is used in the same
        directory with the same name and a '.meta.json' extension.
    :param data_format: (str, default: `None`) format to interpret data
        sources. Will be inferred automatically if not specified.  Valid
        formats are `'auto'`, `'csv'`, `'df'`, `'dict'`, `'excel'`, `'feather'`,
        `'fwf'`, `'hdf5'` (cache file produced during previous training),
        `'html'` (file containing a single HTML `<table>`), `'json'`, `'jsonl'`,
        `'parquet'`, `'pickle'` (pickled Pandas DataFrame), `'sas'`, `'spss'`,
        `'stata'`, `'tsv'`.
    :param experiment_name: (str, default: `'experiment'`) name for
        the experiment.
    :param model_name: (str, default: `'run'`) name of the model that is
        being used.
    :param resume: (bool) If true, continue hyperopt from the state of the previous
        run in the output directory with the same experiment name. If false, will create
        new trials, ignoring any previous state, even if they exist in the output_directory.
        By default, will attempt to resume if there is already an existing experiment with
        the same name, and will create new trials if not.
    :param skip_save_training_description: (bool, default: `False`) disables
        saving the description JSON file.
    :param skip_save_training_statistics: (bool, default: `False`) disables
        saving training statistics JSON file.
    :param skip_save_model: (bool, default: `False`) disables
        saving model weights and hyperparameters each time the model
        improves. By default Ludwig saves model weights after each epoch
        the validation metric improves, but if the model is really big
        that can be time consuming. If you do not want to keep
        the weights and just find out what performance a model can get
        with a set of hyperparameters, use this parameter to skip it,
        but the model will not be loadable later on and the returned model
        will have the weights obtained at the end of training, instead of
        the weights of the epoch with the best validation performance.
    :param skip_save_progress: (bool, default: `False`) disables saving
        progress each epoch. By default Ludwig saves weights and stats
        after each epoch for enabling resuming of training, but if
        the model is really big that can be time consuming and will uses
        twice as much space, use this parameter to skip it, but training
        cannot be resumed later on.
    :param skip_save_log: (bool, default: `False`) disables saving
        TensorBoard logs. By default Ludwig saves logs for the TensorBoard,
        but if it is not needed turning it off can slightly increase the
        overall speed.
    :param skip_save_processed_input: (bool, default: `False`) if input
        dataset is provided it is preprocessed and cached by saving an HDF5
        and JSON files to avoid running the preprocessing again. If this
        parameter is `False`, the HDF5 and JSON file are not saved.
    :param skip_save_unprocessed_output: (bool, default: `False`) by default
        predictions and their probabilities are saved in both raw
        unprocessed numpy files containing tensors and as postprocessed
        CSV files (one for each output feature). If this parameter is True,
        only the CSV ones are saved and the numpy ones are skipped.
    :param skip_save_predictions: (bool, default: `False`) skips saving test
        predictions CSV files.
    :param skip_save_eval_stats: (bool, default: `False`) skips saving test
        statistics JSON file.
    :param skip_save_hyperopt_statistics: (bool, default: `False`) skips saving
        hyperopt stats file.
    :param output_directory: (str, default: `'results'`) the directory that
        will contain the training statistics, TensorBoard logs, the saved
        model and the training progress files.
    :param gpus: (list, default: `None`) list of GPUs that are available
        for training.
    :param gpu_memory_limit: (float: default: `None`) maximum memory fraction
        [0, 1] allowed to allocate per GPU device.
    :param allow_parallel_threads: (bool, default: `True`) allow PyTorch
        to use multithreading parallelism to improve performance at
        the cost of determinism.
    :param callbacks: (list, default: `None`) a list of
        `ludwig.callbacks.Callback` objects that provide hooks into the
        Ludwig pipeline.
    :param backend: (Union[Backend, str]) `Backend` or string name
        of backend to use to execute preprocessing / training steps.
    :param random_seed: (int: default: 42) random seed used for weights
        initialization, splits and any other random function.
    :param hyperopt_log_verbosity: (int: default: 3) controls verbosity of
        ray tune log messages.  Valid values: 0 = silent, 1 = only status updates,
        2 = status and brief trial results, 3 = status and detailed trial results.

    # Return

    :return: (List[dict]) List of results for each trial, ordered by
        descending performance on the target metric.
    """
    from ludwig.hyperopt.execution import get_build_hyperopt_executor, RayTuneExecutor

    # check if config is a path or a dict
    if isinstance(config, str):  # assume path
        with open_file(config, "r") as def_file:
            config_dict = yaml.safe_load(def_file)
    else:
        config_dict = config

    if HYPEROPT not in config:
        raise ValueError("Hyperopt Section not present in config")

    # backwards compatibility
    upgraded_config = upgrade_config_dict_to_latest_version(config_dict)

    # Initialize config object
    config_obj = ModelConfig.from_dict(upgraded_config)

    # Retain pre-merged config for hyperopt schema generation
    premerged_config = copy.deepcopy(upgraded_config)

    # Get full config with defaults
    full_config = config_obj.to_dict()  # TODO (Connor): Refactor to use config object

    hyperopt_config = full_config[HYPEROPT]

    # Explicitly default to a local backend to avoid picking up Ray or Horovod
    # backend from the environment.
    backend = backend or config_dict.get("backend") or "local"
    backend = initialize_backend(backend)

    update_hyperopt_params_with_defaults(hyperopt_config)

    # Check if all features are grid type parameters and log UserWarning if needed
    log_warning_if_all_grid_type_parameters(hyperopt_config)

    # Infer max concurrent trials
    if hyperopt_config[EXECUTOR].get(MAX_CONCURRENT_TRIALS) == AUTO:
        hyperopt_config[EXECUTOR][MAX_CONCURRENT_TRIALS] = backend.max_concurrent_trials(hyperopt_config)
        logger.info(f"Setting max_concurrent_trials to {hyperopt_config[EXECUTOR][MAX_CONCURRENT_TRIALS]}")

    # Print hyperopt config
    logger.info("Hyperopt Config")
    logger.info(pformat(hyperopt_config, indent=4))
    logger.info("\n")

    search_alg = hyperopt_config[SEARCH_ALG]
    executor = hyperopt_config[EXECUTOR]
    parameters = hyperopt_config[PARAMETERS]
    split = hyperopt_config[SPLIT]
    output_feature = hyperopt_config["output_feature"]
    metric = hyperopt_config[METRIC]
    goal = hyperopt_config[GOAL]

    ######################
    # check validity of output_feature / metric/ split combination
    ######################
    splitter = get_splitter(**full_config[PREPROCESSING]["split"])
    if split == TRAINING:
        if training_set is None and not splitter.has_split(0):
            raise ValueError(
                'The data for the specified split for hyperopt "{}" '
                "was not provided, "
                "or the split amount specified in the preprocessing section "
                "of the config is not greater than 0".format(split)
            )
    elif split == VALIDATION:
        if validation_set is None and not splitter.has_split(1):
            raise ValueError(
                'The data for the specified split for hyperopt "{}" '
                "was not provided, "
                "or the split amount specified in the preprocessing section "
                "of the config is not greater than 0".format(split)
            )
    elif split == TEST:
        if test_set is None and not splitter.has_split(2):
            raise ValueError(
                'The data for the specified split for hyperopt "{}" '
                "was not provided, "
                "or the split amount specified in the preprocessing section "
                "of the config is not greater than 0".format(split)
            )
    else:
        raise ValueError(
            'unrecognized hyperopt split "{}". ' "Please provide one of: {}".format(split, {TRAINING, VALIDATION, TEST})
        )
    if output_feature == COMBINED:
        if metric != LOSS:
            raise ValueError('The only valid metric for "combined" output feature is "loss"')
    else:
        output_feature_names = {of[NAME] for of in full_config[OUTPUT_FEATURES]}
        if output_feature not in output_feature_names:
            raise ValueError(
                'The output feature specified for hyperopt "{}" '
                "cannot be found in the config. "
                'Available ones are: {} and "combined"'.format(output_feature, output_feature_names)
            )

    hyperopt_executor = get_build_hyperopt_executor(executor[TYPE])(
        parameters, output_feature, metric, goal, split, search_alg=search_alg, **executor
    )

    if not (
        isinstance(backend, LocalBackend)
        or (isinstance(hyperopt_executor, RayTuneExecutor) and isinstance(backend, RayBackend))
    ):
        raise ValueError(
            "Hyperopt requires using a `local` backend at this time, or " "`ray` backend with `ray` executor."
        )

    for callback in callbacks or []:
        callback.on_hyperopt_init(experiment_name)

    if not should_tune_preprocessing(full_config):
        # preprocessing is not being tuned, so generate it once before starting trials
        for callback in callbacks or []:
            callback.on_hyperopt_preprocessing_start(experiment_name)

        model = LudwigModel(
            config=full_config,
            backend=backend,
            gpus=gpus,
            gpu_memory_limit=gpu_memory_limit,
            allow_parallel_threads=allow_parallel_threads,
            callbacks=callbacks,
        )

        training_set, validation_set, test_set, training_set_metadata = model.preprocess(
            dataset=dataset,
            training_set=training_set,
            validation_set=validation_set,
            test_set=test_set,
            training_set_metadata=training_set_metadata,
            data_format=data_format,
            skip_save_processed_input=skip_save_processed_input,
            random_seed=random_seed,
        )
        dataset = None

        dataset_statistics = generate_dataset_statistics(training_set, validation_set, test_set)

        logger.info("\nDataset Statistics")
        logger.info(tabulate(dataset_statistics, headers="firstrow", tablefmt="fancy_grid"))

        for callback in callbacks or []:
            callback.on_hyperopt_preprocessing_end(experiment_name)

    for callback in callbacks or []:
        callback.on_hyperopt_start(experiment_name)

    hyperopt_results = hyperopt_executor.execute(
        premerged_config,
        dataset=dataset,
        training_set=training_set,
        validation_set=validation_set,
        test_set=test_set,
        training_set_metadata=training_set_metadata,
        data_format=data_format,
        experiment_name=experiment_name,
        model_name=model_name,
        resume=resume,
        skip_save_training_description=skip_save_training_description,
        skip_save_training_statistics=skip_save_training_statistics,
        skip_save_model=skip_save_model,
        skip_save_progress=skip_save_progress,
        skip_save_log=skip_save_log,
        skip_save_processed_input=skip_save_processed_input,
        skip_save_unprocessed_output=skip_save_unprocessed_output,
        skip_save_predictions=skip_save_predictions,
        skip_save_eval_stats=skip_save_eval_stats,
        output_directory=output_directory,
        gpus=gpus,
        gpu_memory_limit=gpu_memory_limit,
        allow_parallel_threads=allow_parallel_threads,
        callbacks=callbacks,
        tune_callbacks=tune_callbacks,
        backend=backend,
        random_seed=random_seed,
        hyperopt_log_verbosity=hyperopt_log_verbosity,
        **kwargs,
    )

    if backend.is_coordinator():
        print_hyperopt_results(hyperopt_results)

        if not skip_save_hyperopt_statistics:
            with backend.storage.artifacts.use_credentials():
                results_directory = os.path.join(output_directory, experiment_name)
                makedirs(results_directory, exist_ok=True)

                hyperopt_stats = {
                    "hyperopt_config": hyperopt_config,
                    "hyperopt_results": [t.to_dict() for t in hyperopt_results.ordered_trials],
                }

                save_hyperopt_stats(hyperopt_stats, results_directory)
                logger.info(f"Hyperopt stats saved to: {results_directory}")

    for callback in callbacks or []:
        callback.on_hyperopt_end(experiment_name)
        callback.on_hyperopt_finish(experiment_name)

    logger.info("Finished hyperopt")

    return hyperopt_results



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import argparse
import random


def split(input_path, output1, output2, split):
    with open(input_path) as file:
        lines = file.readlines()

    random.shuffle(lines)
    split_idx = int(len(lines) * split)

    with open(output1, "w") as f:
        for line in lines[:split_idx]:
            line = line if line.endswith("\n") else line + "\n"
            f.write(line)

    with open(output2, "w") as f:
        for line in lines[split_idx:]:
            line = line if line.endswith("\n") else line + "\n"
            f.write(line)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Split a file based on its lines")

    parser.add_argument("-i", "--input", required=True, help="input file names")
    parser.add_argument("-o1", "--output1", required=True, help="output 1 file name")
    parser.add_argument("-o2", "--output2", required=True, help="output 2 file name")
    parser.add_argument("-s", "--split", required=True, type=float, default=0.8, help="percentage of the split")

    args = parser.parse_args()

    split(args.input, args.output1, args.output2, args.split)



import dataclasses
from dataclasses import dataclass
from typing import Dict, Union

from ludwig.utils.data_utils import flatten_dict


@dataclass
class DeviceUsageMetrics:
    # Max CUDA memory utilization of the code block.
    max_memory_used: float

    # Average CUDA memory utilization of the code block.
    average_memory_used: float


@dataclass
class SystemResourceMetrics:
    # Name of the code block/function to be profiled.
    code_block_tag: str

    # Name of the CPU that the code ran on.
    cpu_name: str

    # CPU architecture that the code ran on.
    cpu_architecture: str

    # Number of CPUs on the machine.
    num_cpu: int

    # Total CPU memory size.
    total_cpu_memory_size: float

    # Ludwig version in the environment.
    ludwig_version: str

    # Total execution time of the code block.
    total_execution_time: float

    # The change in disk memory before and after the code block ran.
    disk_footprint: float

    # Max CPU utilization of the code block.
    max_cpu_utilization: float

    # Max CPU memory (RAM) utilization of the code block.
    max_cpu_memory_usage: float

    # Min system-wide CPU memory available (how much physical memory is left).
    min_global_cpu_memory_available: float

    # Max system-wide CPU utilization.
    max_global_cpu_utilization: float

    # Average CPU utilization of the code block.
    average_cpu_utilization: float

    # Average CPU memory (RAM) utilization of the code block.
    average_cpu_memory_usage: float

    # Average system-wide CPU memory available (how much physical memory is left).
    average_global_cpu_memory_available: float

    # Average system-wide CPU utilization.
    average_global_cpu_utilization: float

    # Per device usage. Dictionary containing max and average memory used per device.
    device_usage: Dict[str, DeviceUsageMetrics]


@dataclass
class TorchProfilerMetrics:
    # Time taken by torch ops to execute on the CPU.
    torch_cpu_time: float

    # Time taken by torch ops to execute on CUDA devices.
    torch_cuda_time: float

    # Number of out of memory events.
    num_oom_events: int

    # Per device usage by torch ops. Dictionary containing max and average memory used per device.
    device_usage: Dict[str, DeviceUsageMetrics]


def profiler_dataclass_to_flat_dict(data: Union[SystemResourceMetrics, TorchProfilerMetrics]) -> Dict:
    """Returns a flat dictionary representation, with the device_usage key removed."""
    nested_dict = dataclasses.asdict(data)
    nested_dict[""] = nested_dict.pop("device_usage")
    return flatten_dict(nested_dict, sep="")



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import random

import numpy as np
import pytest
import torch

from ludwig.data.dataset_synthesizer import build_vocab
from ludwig.encoders.base import Encoder
from ludwig.encoders.image.base import MLPMixerEncoder, Stacked2DCNN
from ludwig.encoders.sequence_encoders import (
    ParallelCNN,
    SequenceEmbedEncoder,
    StackedCNN,
    StackedCNNRNN,
    StackedParallelCNN,
    StackedRNN,
)
from ludwig.utils.torch_utils import get_torch_device
from tests.integration_tests.parameter_update_utils import check_module_parameters_updated

DROPOUT = 0.5
DEVICE = get_torch_device()
RANDOM_SEED = 1919


def create_encoder(encoder_type, **encoder_kwargs):
    encoder = encoder_type(**encoder_kwargs)
    return encoder


def _generate_image(image_size):
    return np.random.randint(0, 1, image_size).astype(np.float32)


def generate_images(image_size, num_images):
    return np.array([_generate_image(image_size) for _ in range(num_images)])


def _generate_sentence(vocab_size, max_len):
    sentence = np.zeros(max_len, dtype=np.int32)
    random_length = random.randint(1, max_len)
    sentence[:random_length] = [random.randint(0, vocab_size - 1) for _ in range(random_length)]

    return sentence


def generate_random_sentences(num_sentences=10, max_len=10, vocab_size=10):
    # Generate some random text
    vocab = build_vocab(vocab_size)

    text = np.array([_generate_sentence(vocab_size, max_len) for _ in range(num_sentences)])

    return text, vocab


def encoder_test(
    encoder,
    input_data,
    output_dtype,
    output_shape,
    output_data=None,
):
    """Helper method to test different kinds of encoders.

    :param encoder: encoder object
    :param input_data: data to encode
    :param output_dtype: expected data type of the output (optional)
    :param output_shape: expected shape of the encoder output (optional)
    :param output_data: expected output data (optional)
    :return: returns the encoder object for the caller to run extra checks
    """
    encoder = encoder.to(DEVICE)

    # Run the encoder
    input_data = torch.from_numpy(input_data).to(DEVICE)

    hidden = encoder(input_data)["encoder_output"]

    # Check output shape and type
    assert hidden.dtype == output_dtype
    assert list(hidden.shape) == output_shape

    if output_data is not None:
        # todo the hidden output is actually a tensor. May need modification
        assert np.allclose(hidden, output_data)


def test_image_encoders_stacked_2dcnn():
    # make repeatable
    np.random.seed(RANDOM_SEED)
    torch.manual_seed(RANDOM_SEED)

    # Test the resnet encoder for images
    encoder_kwargs = {"num_conv_layers": 2, "num_filters": 16, "output_size": 28, "dropout": DROPOUT}
    image_size = (3, 10, 10)

    encoder = create_encoder(
        Stacked2DCNN, num_channels=image_size[0], height=image_size[1], width=image_size[2], **encoder_kwargs
    )

    assert encoder is not None
    assert encoder.conv_stack_2d is not None
    assert list(encoder.conv_stack_2d.output_shape) == [32, 1, 1]
    assert len(encoder.fc_stack.layers) == 1
    assert encoder.conv_stack_2d.layers[0]["pool_kernel_size"] == 2
    assert encoder.conv_stack_2d.layers[0]["stride"] == 1
    assert encoder.conv_stack_2d.layers[0]["pool_stride"] == 2
    assert encoder.conv_stack_2d.layers[0]["norm"] is None
    assert encoder.conv_stack_2d.layers[0]["activation"] == "relu"
    assert encoder.conv_stack_2d.layers[0]["dropout"] == 0

    output_shape = [1, 28]
    input_image = generate_images(image_size, 1)

    encoder_test(
        encoder=encoder, input_data=input_image, output_dtype=torch.float32, output_shape=output_shape, output_data=None
    )

    output_shape = [5, 28]
    input_images = generate_images(image_size, 5)

    encoder_test(
        encoder=encoder,
        input_data=input_images,
        output_dtype=torch.float32,
        output_shape=output_shape,
        output_data=None,
    )

    # test for parameter updates
    # generate tensors for parameter update test
    target = torch.rand(output_shape)
    image_tensor = torch.rand(input_image.shape)

    # check for parameter updates
    fpc, tpc, upc, not_updated = check_module_parameters_updated(encoder, (image_tensor,), target)
    assert upc == tpc, (
        f"Not all trainable parameters updated.  Parameters not updated: {not_updated}."
        f"  Module structure\n{encoder}"
    )


def test_image_encoders_mlpmixer():
    # make repeatable
    np.random.seed(RANDOM_SEED)
    torch.manual_seed(RANDOM_SEED)

    # Test the resnet encoder for images
    encoder_kwargs = {
        "patch_size": 5,
        "embed_size": 8,
        "token_size": 32,
        "channel_dim": 16,
        "num_layers": 2,
        "dropout": DROPOUT,
    }
    image_size = (3, 10, 10)

    output_shape = [1, 8]
    input_image = generate_images(image_size, 1)

    encoder = create_encoder(
        MLPMixerEncoder, num_channels=image_size[0], height=image_size[1], width=image_size[2], **encoder_kwargs
    )
    encoder_test(
        encoder=encoder, input_data=input_image, output_dtype=torch.float32, output_shape=output_shape, output_data=None
    )

    output_shape = [5, 8]
    input_images = generate_images(image_size, 5)

    encoder_test(
        encoder=encoder,
        input_data=input_images,
        output_dtype=torch.float32,
        output_shape=output_shape,
        output_data=None,
    )

    assert encoder is not None
    assert encoder.mlp_mixer.__class__.__name__ == "MLPMixer"
    assert len(encoder.mlp_mixer.mixer_blocks) == 2
    assert list(encoder.mlp_mixer.mixer_blocks[0].mlp1.output_shape) == [4]
    assert encoder.mlp_mixer.patch_conv.__class__.__name__ == "Conv2d"
    assert encoder.mlp_mixer.patch_conv.kernel_size == (5, 5)

    # test for parameter updates
    # generate tensors for parameter update test
    target = torch.rand(output_shape)
    image_tensor = torch.rand(input_image.shape)

    # check for parameter updates
    fpc, tpc, upc, not_updated = check_module_parameters_updated(encoder, (image_tensor,), target)
    assert upc == tpc, (
        f"Not all trainable parameters updated.  Parameters not updated: {not_updated}."
        f"  Module structure\n{encoder}"
    )


def test_sequence_encoder_embed():
    num_sentences = 4
    embedding_size = 5
    max_len = 6

    # make repeatable
    np.random.seed(RANDOM_SEED)
    torch.manual_seed(RANDOM_SEED)

    # Generate data
    text, vocab = generate_random_sentences(
        num_sentences=num_sentences,
        max_len=max_len,
    )

    encoder_kwargs = {"embedding_size": embedding_size, "vocab": vocab}

    # Different values for reduce_output and the corresponding expected size
    reduce_outputs = ["sum", None, "concat"]
    output_shapes = [
        [num_sentences, embedding_size],
        [num_sentences, max_len, embedding_size],
        [num_sentences, max_len * embedding_size],
    ]

    for reduce_output, output_shape in zip(reduce_outputs, output_shapes):
        for trainable in [True, False]:
            encoder_kwargs["reduce_output"] = reduce_output
            encoder_kwargs["embeddings_trainable"] = trainable
            encoder_kwargs["dropout"] = DROPOUT
            encoder = create_encoder(SequenceEmbedEncoder, max_sequence_length=max_len, **encoder_kwargs)

            encoder_test(
                encoder=encoder,
                input_data=text,
                output_dtype=torch.float32,
                output_shape=output_shape,
                output_data=None,
            )

            assert encoder.embed_sequence.dropout is not None

            # test for parameter updates
            # generate tensors for parameter update test
            target = torch.rand(output_shape)

            # check for parameter updates
            fpc, tpc, upc, not_updated = check_module_parameters_updated(
                encoder, (torch.tensor(text, dtype=torch.int32),), target
            )
            assert upc == tpc, (
                f"Not all trainable parameters updated.  Parameters not updated: {not_updated}."
                f"  Module structure\n{encoder}"
            )


@pytest.mark.parametrize("encoder_type", [ParallelCNN, StackedCNN, StackedParallelCNN, StackedRNN, StackedCNNRNN])
@pytest.mark.parametrize("trainable", [True, False])
@pytest.mark.parametrize("reduce_output", ["sum", "max"])
def test_sequence_encoders(encoder_type: Encoder, trainable: bool, reduce_output: str):
    num_sentences = 4
    embedding_size = 5
    max_len = 7
    output_size = 3

    # make repeatable
    np.random.seed(RANDOM_SEED)
    torch.manual_seed(RANDOM_SEED)

    # Generate data
    text, vocab = generate_random_sentences(
        num_sentences=num_sentences,
        max_len=max_len,
    )

    encoder_kwargs = {
        "embedding_size": embedding_size,
        "vocab": vocab,
        "output_size": output_size,
        "num_fc_layers": 1,
        "filter_size": 3,
        "num_filters": 8,
        "state_size": output_size,
    }

    # todo figure out the output size for parallel 1d conv
    output_shape = [num_sentences, output_size]

    encoder_kwargs["embeddings_trainable"] = trainable
    encoder_kwargs["dropout"] = DROPOUT
    encoder_kwargs["recurrent_dropout"] = DROPOUT
    encoder_kwargs["fc_dropout"] = DROPOUT
    encoder_kwargs["reduce_output"] = reduce_output
    encoder = create_encoder(encoder_type, max_sequence_length=max_len, **encoder_kwargs)

    encoder_test(
        encoder=encoder, input_data=text, output_dtype=torch.float32, output_shape=output_shape, output_data=None
    )

    assert isinstance(encoder, encoder_type)

    # test for parameter updates
    # generate tensors for parameter update test
    target = torch.rand(output_shape)

    # check for parameter updates
    fpc, tpc, upc, not_updated = check_module_parameters_updated(
        encoder, (torch.tensor(text, dtype=torch.int32),), target
    )

    if trainable:
        assert fpc == 0, "Embedding layer expected to be trainable but found to be frozen"
    else:
        assert fpc == 1, "Embedding layer expected to be frozen, but found to be trainable."

    # for given random seed and configuration and non-zero dropout updated parameter counts
    # could take on different values
    assert (upc == tpc) or (upc == 0), (
        f"Not all trainable parameters updated.  Parameters not updated: {not_updated}."
        f"  Module structure\n{encoder}"
    )



from abc import ABC
from typing import Optional, Union

import torch
from packaging.version import parse as parse_version

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import DEFAULT_BATCH_SIZE, LOSS, MAX_POSSIBLE_BATCH_SIZE, MODEL_ECD, MODEL_GBM, TRAINING
from ludwig.error import ConfigValidationError
from ludwig.schema import utils as schema_utils
from ludwig.schema.lr_scheduler import LRSchedulerConfig, LRSchedulerDataclassField
from ludwig.schema.metadata import TRAINER_METADATA
from ludwig.schema.optimizers import (
    BaseOptimizerConfig,
    GradientClippingConfig,
    GradientClippingDataclassField,
    OptimizerDataclassField,
)
from ludwig.schema.utils import ludwig_dataclass
from ludwig.utils.registry import Registry

_torch_200 = parse_version(torch.__version__) >= parse_version("2.0")


trainer_schema_registry = Registry()


@DeveloperAPI
def register_trainer_schema(model_type: str):
    def wrap(trainer_config: BaseTrainerConfig):
        trainer_schema_registry[model_type] = trainer_config
        return trainer_config

    return wrap


@DeveloperAPI
@ludwig_dataclass
class BaseTrainerConfig(schema_utils.BaseMarshmallowConfig, ABC):
    """Common trainer parameter values."""

    validation_field: str = schema_utils.String(
        default=None,
        allow_none=True,
        description="The field for which the `validation_metric` is used for validation-related mechanics like early "
        "stopping, parameter change plateaus, as well as what hyperparameter optimization uses to determine the best "
        "trial. If unset (default), the first output feature is used. If explicitly specified, neither "
        "`validation_field` nor `validation_metric` are overwritten.",
    )

    validation_metric: str = schema_utils.String(
        default=None,
        allow_none=True,
        description=(
            "Metric from `validation_field` that is used. If validation_field is not explicitly specified, this is "
            "overwritten to be the first output feature type's `default_validation_metric`, consistent with "
            "validation_field. If the validation_metric is specified, then we will use the first output feature that "
            "produces this metric as the `validation_field`."
        ),
    )

    early_stop: int = schema_utils.IntegerRange(
        default=5,
        min=-1,
        description=(
            "Number of consecutive rounds of evaluation without any improvement on the `validation_metric` that "
            "triggers training to stop. Can be set to -1, which disables early stopping entirely."
        ),
    )


@DeveloperAPI
@register_trainer_schema(MODEL_ECD)
@ludwig_dataclass
class ECDTrainerConfig(BaseTrainerConfig):
    """Dataclass that configures most of the hyperparameters used for ECD model training."""

    def __post_init__(self):
        if self.compile and not _torch_200:
            raise ConfigValidationError(
                "Trainer param `compile: true` requires PyTorch 2.0.0 or higher. Please upgrade PyTorch and try again."
            )

    learning_rate: Union[float, str] = schema_utils.OneOfOptionsField(
        default=0.001,
        allow_none=False,
        description=(
            "Controls how much to change the model in response to the estimated error each time the model weights are "
            "updated. If 'auto', the optimal learning rate is estimated by choosing the learning rate that produces "
            "the smallest non-diverging gradient update."
        ),
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["learning_rate"],
        field_options=[
            schema_utils.FloatRange(default=0.001, allow_none=False, min=0, max=1),
            schema_utils.StringOptions(options=["auto"], default="auto", allow_none=False),
        ],
    )

    learning_rate_scheduler: LRSchedulerConfig = LRSchedulerDataclassField(
        description="Parameter values for learning rate scheduler.",
        default=None,
    )

    epochs: int = schema_utils.PositiveInteger(
        default=100,
        description="Number of epochs the algorithm is intended to be run over. Overridden if `train_steps` is set",
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["epochs"],
    )

    checkpoints_per_epoch: int = schema_utils.NonNegativeInteger(
        default=0,
        description=(
            "Number of checkpoints per epoch. For example, 2 -> checkpoints are written every half of an epoch. Note "
            "that it is invalid to specify both non-zero `steps_per_checkpoint` and non-zero `checkpoints_per_epoch`."
        ),
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["checkpoints_per_epoch"],
    )

    train_steps: int = schema_utils.PositiveInteger(
        default=None,
        allow_none=True,
        description=(
            "Maximum number of training steps the algorithm is intended to be run over. Unset by default. "
            "If set, will override `epochs` and if left unset then `epochs` is used to determine training length."
        ),
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["train_steps"],
    )

    steps_per_checkpoint: int = schema_utils.NonNegativeInteger(
        default=0,
        description=(
            "How often the model is checkpointed. Also dictates maximum evaluation frequency. If 0 the model is "
            "checkpointed after every epoch."
        ),
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["steps_per_checkpoint"],
    )

    batch_size: Union[int, str] = schema_utils.OneOfOptionsField(
        default=DEFAULT_BATCH_SIZE,
        allow_none=False,
        description=(
            "The number of training examples utilized in one training step of the model. If ’auto’, the "
            "batch size that maximized training throughput (samples / sec) will be used. For CPU training, the "
            "tuned batch size is capped at 128 as throughput benefits of large batch sizes are less noticeable without "
            "a GPU."
        ),
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["batch_size"],
        field_options=[
            schema_utils.PositiveInteger(default=128, description="", allow_none=False),
            schema_utils.StringOptions(options=["auto"], default="auto", allow_none=False),
        ],
    )

    max_batch_size: int = schema_utils.PositiveInteger(
        default=MAX_POSSIBLE_BATCH_SIZE,
        allow_none=True,
        description=(
            "Auto batch size tuning and increasing batch size on plateau will be capped at this value. The default "
            "value is 2^40."
        ),
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["max_batch_size"],
    )

    early_stop: int = schema_utils.IntegerRange(
        default=5,
        min=-1,
        description=(
            "Number of consecutive rounds of evaluation without any improvement on the `validation_metric` that "
            "triggers training to stop. Can be set to -1, which disables early stopping entirely."
        ),
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["early_stop"],
    )

    eval_batch_size: Union[None, int, str] = schema_utils.OneOfOptionsField(
        default=None,
        allow_none=True,
        description=(
            "Size of batch to pass to the model for evaluation. If it is `0` or `None`, the same value of `batch_size` "
            "is used. This is useful to speedup evaluation with a much bigger batch size than training, if enough "
            "memory is available. If ’auto’, the biggest batch size (power of 2) that can fit in memory will be used."
        ),
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["eval_batch_size"],
        field_options=[
            schema_utils.PositiveInteger(default=128, description="", allow_none=False),
            schema_utils.StringOptions(options=["auto"], default="auto", allow_none=False),
        ],
    )

    evaluate_training_set: bool = schema_utils.Boolean(
        default=False,
        description=(
            "Whether to evaluate on the entire training set during evaluation. By default, training metrics will be "
            "computed at the end of each training step, and accumulated up to the evaluation phase. In practice, "
            "computing training set metrics during training is up to 30% faster than running a separate evaluation "
            "pass over the training set, but results in more noisy training metrics, particularly during the earlier "
            "epochs. It's recommended to only set this to True if you need very exact training set metrics, and are "
            "willing to pay a significant performance penalty for them."
        ),
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["evaluate_training_set"],
    )

    validation_field: str = schema_utils.String(
        default=None,
        allow_none=True,
        description="The field for which the `validation_metric` is used for validation-related mechanics like early "
        "stopping, parameter change plateaus, as well as what hyperparameter optimization uses to determine the best "
        "trial. If unset (default), the first output feature is used. If explicitly specified, neither "
        "`validation_field` nor `validation_metric` are overwritten.",
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["validation_field"],
    )

    validation_metric: str = schema_utils.String(
        default=None,
        allow_none=True,
        description=(
            "Metric from `validation_field` that is used. If validation_field is not explicitly specified, this is "
            "overwritten to be the first output feature type's `default_validation_metric`, consistent with "
            "validation_field. If the validation_metric is specified, then we will use the first output feature that "
            "produces this metric as the `validation_field`."
        ),
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["validation_metric"],
    )

    optimizer: BaseOptimizerConfig = OptimizerDataclassField(
        default="adam",
        description=(
            "Optimizer type and its parameters. The optimizer is responsble for applying the gradients computed "
            "from the loss during backpropagation as updates to the model weights."
        ),
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["optimizer"],
    )

    regularization_type: Optional[str] = schema_utils.RegularizerOptions(
        default="l2",
        allow_none=True,
        description="Type of regularization.",
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["regularization_type"],
    )

    regularization_lambda: float = schema_utils.FloatRange(
        default=0.0,
        min=0,
        max=1,
        description="Strength of the regularization.",
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["regularization_lambda"],
    )

    should_shuffle: bool = schema_utils.Boolean(
        default=True,
        description="Whether to shuffle batches during training when true.",
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["should_shuffle"],
    )

    increase_batch_size_on_plateau: int = schema_utils.NonNegativeInteger(
        default=0,
        description="The number of times to increase the batch size on a plateau.",
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["increase_batch_size_on_plateau"],
    )

    increase_batch_size_on_plateau_patience: int = schema_utils.NonNegativeInteger(
        default=5,
        description="How many epochs to wait for before increasing the batch size.",
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["increase_batch_size_on_plateau_patience"],
    )

    increase_batch_size_on_plateau_rate: float = schema_utils.NonNegativeFloat(
        default=2.0,
        description="Rate at which the batch size increases.",
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["increase_batch_size_on_plateau_rate"],
    )

    increase_batch_size_eval_metric: str = schema_utils.String(
        default=LOSS,
        description="Which metric to listen on for increasing the batch size.",
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["increase_batch_size_eval_metric"],
    )

    increase_batch_size_eval_split: str = schema_utils.String(
        default=TRAINING,
        description="Which dataset split to listen on for increasing the batch size.",
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["increase_batch_size_eval_split"],
    )

    gradient_clipping: Optional[GradientClippingConfig] = GradientClippingDataclassField(
        description="Parameter values for gradient clipping.",
        default={},
    )

    learning_rate_scaling: str = schema_utils.StringOptions(
        ["constant", "sqrt", "linear"],
        default="linear",
        description="Scale by which to increase the learning rate as the number of distributed workers increases. "
        "Traditionally the learning rate is scaled linearly with the number of workers to reflect the "
        "proportion by"
        " which the effective batch size is increased. For very large batch sizes, a softer square-root "
        "scale can "
        "sometimes lead to better model performance. If the learning rate is hand-tuned for a given "
        "number of "
        "workers, setting this value to constant can be used to disable scale-up.",
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["learning_rate_scaling"],
    )

    bucketing_field: str = schema_utils.String(
        default=None,
        allow_none=True,
        description="Feature to use for bucketing datapoints",
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["bucketing_field"],
    )

    use_mixed_precision: bool = schema_utils.Boolean(
        default=False,
        description="Enable automatic mixed-precision (AMP) during training.",
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["use_mixed_precision"],
    )

    compile: bool = schema_utils.Boolean(
        default=False,
        description="Whether to compile the model before training.",
        parameter_metadata=TRAINER_METADATA[MODEL_ECD]["compile"],
    )


@DeveloperAPI
@register_trainer_schema(MODEL_GBM)
@ludwig_dataclass
class GBMTrainerConfig(BaseTrainerConfig):
    """Dataclass that configures most of the hyperparameters used for GBM model training."""

    # NOTE: Overwritten here since GBM performs better with a different default learning rate.
    learning_rate: Union[float, str] = schema_utils.NonNegativeFloat(
        default=0.03,
        allow_none=False,
        description=(
            "Controls how much to change the model in response to the estimated error each time the model weights are "
            "updated."
        ),
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["learning_rate"],
    )

    early_stop: int = schema_utils.IntegerRange(
        default=5,
        min=-1,
        description=(
            "Number of consecutive rounds of evaluation without any improvement on the `validation_metric` that "
            "triggers training to stop. Can be set to -1, which disables early stopping entirely."
        ),
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["early_stop"],
    )

    # LightGBM Learning Control params
    max_depth: int = schema_utils.Integer(
        default=18,
        description="Maximum depth of a tree in the GBM trainer. A negative value means no limit.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["max_depth"],
    )

    drop_rate: float = schema_utils.FloatRange(
        default=0.1,
        min=0,
        max=1,
        description="Dropout rate for the GBM trainer. Used only with boosting_type 'dart'.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["drop_rate"],
    )

    # NOTE: Overwritten here to provide a default value. In many places, we fall back to eval_batch_size if batch_size
    # is not specified. GBM does not have a value for batch_size, so we need to specify eval_batch_size here.
    eval_batch_size: Union[None, int, str] = schema_utils.PositiveInteger(
        default=1048576,
        description="Size of batch to pass to the model for evaluation.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["eval_batch_size"],
    )

    evaluate_training_set: bool = schema_utils.Boolean(
        default=False,
        description=(
            "Whether to evaluate on the entire training set during evaluation. By default, training metrics will be "
            "computed at the end of each training step, and accumulated up to the evaluation phase. In practice, "
            "computing training set metrics during training is up to 30% faster than running a separate evaluation "
            "pass over the training set, but results in more noisy training metrics, particularly during the earlier "
            "epochs. It's recommended to only set this to True if you need very exact training set metrics, and are "
            "willing to pay a significant performance penalty for them."
        ),
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["evaluate_training_set"],
    )

    # TODO(#1673): Need some more logic here for validating against output features
    validation_field: str = schema_utils.String(
        default=None,
        allow_none=True,
        description="First output feature, by default it is set as the same field of the first output feature.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["validation_field"],
    )

    validation_metric: str = schema_utils.String(
        default=None,
        allow_none=True,
        description=(
            "Metric used on `validation_field`, set by default to the "
            "output feature type's `default_validation_metric`."
        ),
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["validation_metric"],
    )

    tree_learner: str = schema_utils.StringOptions(
        ["serial", "feature", "data", "voting"],
        allow_none=False,
        default="serial",
        description="Type of tree learner to use with GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["tree_learner"],
    )

    # LightGBM core parameters (https://lightgbm.readthedocs.io/en/latest/Parameters.html)
    boosting_type: str = schema_utils.StringOptions(
        # TODO: Re-enable "goss" when supported: https://github.com/ludwig-ai/ludwig/issues/2988
        ["gbdt", "dart"],
        allow_none=False,
        default="gbdt",
        description="Type of boosting algorithm to use with GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["boosting_type"],
    )

    boosting_rounds_per_checkpoint: int = schema_utils.PositiveInteger(
        default=50,
        description="Number of boosting rounds per checkpoint / evaluation round.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["boosting_rounds_per_checkpoint"],
    )

    num_boost_round: int = schema_utils.PositiveInteger(
        default=1000,
        description="Number of boosting rounds to perform with GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["num_boost_round"],
    )

    num_leaves: int = schema_utils.PositiveInteger(
        default=82,
        description="Number of leaves to use in the tree with GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["num_leaves"],
    )

    min_data_in_leaf: int = schema_utils.NonNegativeInteger(
        default=20,
        description="Minimum number of data points in a leaf with GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["min_data_in_leaf"],
    )

    min_sum_hessian_in_leaf: float = schema_utils.NonNegativeFloat(
        default=1e-3,
        description="Minimum sum of hessians in a leaf with GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["min_sum_hessian_in_leaf"],
    )

    bagging_fraction: float = schema_utils.FloatRange(
        default=0.8,
        min=0,
        max=1,
        description="Fraction of data to use for bagging with GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["bagging_fraction"],
    )

    pos_bagging_fraction: float = schema_utils.FloatRange(
        default=1.0,
        min=0,
        max=1,
        description="Fraction of positive data to use for bagging with GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["pos_bagging_fraction"],
    )

    neg_bagging_fraction: float = schema_utils.FloatRange(
        default=1.0,
        min=0,
        max=1,
        description="Fraction of negative data to use for bagging with GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["neg_bagging_fraction"],
    )

    bagging_freq: int = schema_utils.NonNegativeInteger(
        default=1,
        description="Frequency of bagging with GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["bagging_freq"],
    )

    bagging_seed: int = schema_utils.Integer(
        default=3,
        description="Random seed for bagging with GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["bagging_seed"],
    )

    feature_fraction: float = schema_utils.FloatRange(
        default=0.75,
        min=0,
        max=1,
        description="Fraction of features to use in the GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["feature_fraction"],
    )

    feature_fraction_bynode: float = schema_utils.FloatRange(
        default=1.0,
        min=0,
        max=1,
        description="Fraction of features to use for each tree node with GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["feature_fraction_bynode"],
    )

    feature_fraction_seed: int = schema_utils.Integer(
        default=2,
        description="Random seed for feature fraction with GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["feature_fraction_seed"],
    )

    extra_trees: bool = schema_utils.Boolean(
        default=False,
        description="Whether to use extremely randomized trees in the GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["extra_trees"],
    )

    extra_seed: int = schema_utils.Integer(
        default=6,
        description="Random seed for extremely randomized trees in the GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["extra_seed"],
    )

    max_delta_step: float = schema_utils.FloatRange(
        default=0.0,
        min=0,
        max=1,
        description=(
            "Used to limit the max output of tree leaves in the GBM trainer. A negative value means no constraint."
        ),
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["max_delta_step"],
    )

    lambda_l1: float = schema_utils.NonNegativeFloat(
        default=0.25,
        description="L1 regularization factor for the GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["lambda_l1"],
    )

    lambda_l2: float = schema_utils.NonNegativeFloat(
        default=0.2,
        description="L2 regularization factor for the GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["lambda_l2"],
    )

    linear_lambda: float = schema_utils.NonNegativeFloat(
        default=0.0,
        description="Linear tree regularization in the GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["linear_lambda"],
    )

    min_gain_to_split: float = schema_utils.NonNegativeFloat(
        default=0.03,
        description="Minimum gain to split a leaf in the GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["min_gain_to_split"],
    )

    max_drop: int = schema_utils.Integer(
        default=50,
        description=(
            "Maximum number of dropped trees during one boosting iteration. "
            "Used only with boosting_type 'dart'. A negative value means no limit."
        ),
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["max_drop"],
    )

    skip_drop: float = schema_utils.FloatRange(
        default=0.5,
        min=0,
        max=1,
        description=(
            "Probability of skipping the dropout during one boosting iteration. Used only with boosting_type 'dart'."
        ),
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["skip_drop"],
    )

    xgboost_dart_mode: bool = schema_utils.Boolean(
        default=False,
        description="Whether to use xgboost dart mode in the GBM trainer. Used only with boosting_type 'dart'.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["xgboost_dart_mode"],
    )

    uniform_drop: bool = schema_utils.Boolean(
        default=False,
        description="Whether to use uniform dropout in the GBM trainer. Used only with boosting_type 'dart'.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["uniform_drop"],
    )

    drop_seed: int = schema_utils.Integer(
        default=4,
        description="Random seed to choose dropping models in the GBM trainer. Used only with boosting_type 'dart'.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["drop_seed"],
    )

    top_rate: float = schema_utils.FloatRange(
        default=0.2,
        min=0,
        max=1,
        description="The retain ratio of large gradient data in the GBM trainer. Used only with boosting_type 'goss'.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["top_rate"],
    )

    other_rate: float = schema_utils.FloatRange(
        default=0.1,
        min=0,
        max=1,
        description="The retain ratio of small gradient data in the GBM trainer. Used only with boosting_type 'goss'.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["other_rate"],
    )

    min_data_per_group: int = schema_utils.PositiveInteger(
        default=100,
        description="Minimum number of data points per categorical group for the GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["min_data_per_group"],
    )

    max_cat_threshold: int = schema_utils.PositiveInteger(
        default=32,
        description="Number of split points considered for categorical features for the GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["max_cat_threshold"],
    )

    cat_l2: float = schema_utils.NonNegativeFloat(
        default=10.0,
        description="L2 regularization factor for categorical split in the GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["cat_l2"],
    )

    cat_smooth: float = schema_utils.NonNegativeFloat(
        default=10.0,
        description="Smoothing factor for categorical split in the GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["cat_smooth"],
    )

    max_cat_to_onehot: int = schema_utils.PositiveInteger(
        default=4,
        description="Maximum categorical cardinality required before one-hot encoding in the GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["max_cat_to_onehot"],
    )

    cegb_tradeoff: float = schema_utils.NonNegativeFloat(
        default=1.0,
        description="Cost-effective gradient boosting multiplier for all penalties in the GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["cegb_tradeoff"],
    )

    cegb_penalty_split: float = schema_utils.NonNegativeFloat(
        default=0.0,
        description="Cost-effective gradient boosting penalty for splitting a node in the GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["cegb_penalty_split"],
    )

    path_smooth: float = schema_utils.NonNegativeFloat(
        default=0.0,
        description="Smoothing factor applied to tree nodes in the GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["path_smooth"],
    )

    verbose: int = schema_utils.IntegerOptions(
        options=list(range(-1, 3)),
        allow_none=False,
        default=-1,
        description="Verbosity level for GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["verbose"],
    )

    # LightGBM IO params
    max_bin: int = schema_utils.PositiveInteger(
        default=255,
        description="Maximum number of bins to use for discretizing features with GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["max_bin"],
    )

    feature_pre_filter: bool = schema_utils.Boolean(
        default=True,
        description="Whether to ignore features that are unsplittable based on min_data_in_leaf in the GBM trainer.",
        parameter_metadata=TRAINER_METADATA[MODEL_GBM]["feature_pre_filter"],
    )


@DeveloperAPI
def get_model_type_jsonschema(model_type: str = MODEL_ECD):
    enum = [MODEL_ECD]
    if model_type == MODEL_GBM:
        enum = [MODEL_GBM]

    return {
        "type": "string",
        "enum": enum,
        "default": MODEL_ECD,
        "title": "model_type",
        "description": "Select the model type.",
    }


@DeveloperAPI
def get_trainer_jsonschema(model_type: str):
    trainer_cls = trainer_schema_registry[model_type]
    props = schema_utils.unload_jsonschema_from_marshmallow_class(trainer_cls)["properties"]

    return {
        "type": "object",
        "properties": props,
        "title": "trainer_options",
        "additionalProperties": False,
        "description": "Schema for trainer determined by Model Type",
    }


@DeveloperAPI
class ECDTrainerField(schema_utils.DictMarshmallowField):
    def __init__(self):
        super().__init__(ECDTrainerConfig)

    @staticmethod
    def _jsonschema_type_mapping():
        return get_trainer_jsonschema(MODEL_ECD)


@DeveloperAPI
class GBMTrainerField(schema_utils.DictMarshmallowField):
    def __init__(self):
        super().__init__(GBMTrainerConfig)

    @staticmethod
    def _jsonschema_type_mapping():
        return get_trainer_jsonschema(MODEL_GBM)



from ludwig.utils.registry import DEFAULT_KEYS, Registry

trainers_registry = Registry()
ray_trainers_registry = Registry()


def register_trainer(model_type: str, default=False):
    """Register a trainer class that supports training the given model types.

    Using default=True will make the trainer the default trainer for the model type.

    Args:
        model_type: The model_type which dictates the trainer type to use.
        default: Whether the trainer should be the default trainer for the model type.
    """

    def wrap(cls):
        trainers_registry[model_type] = cls
        if default:
            if DEFAULT_KEYS[0] in trainers_registry:
                raise ValueError(f"Default trainer already registered for model type {model_type}")
            for key in DEFAULT_KEYS:
                trainers_registry[key] = cls
        return cls

    return wrap


def register_ray_trainer(model_type: str, default=False):
    """Register a trainer class that supports training the given model types with Ray backend.

    Using default=True will make the trainer the default trainer for the model type.

    Args:
        model_type: The model_type which dictates the trainer type to use.
        default: Whether the trainer should be the default trainer for the model type.
    """

    def wrap(cls):
        ray_trainers_registry[model_type] = cls
        if default:
            if DEFAULT_KEYS[0] in ray_trainers_registry:
                raise ValueError(f"Default trainer already registered for model type {model_type}")
            for key in DEFAULT_KEYS:
                ray_trainers_registry[key] = cls
        return cls

    return wrap



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from copy import deepcopy
from typing import Dict, List, Optional

import torch
from torch.nn import Dropout, Linear, ModuleList

from ludwig.modules.normalization_modules import create_norm_layer
from ludwig.utils.torch_utils import activations, initializer_registry, LudwigModule

logger = logging.getLogger(__name__)


class FCLayer(LudwigModule):
    """A torch.nn.Linear wrapper that declares input and output shapes, and enables the customization of:

    1. how weights and biases are initialized
    2. normalization (layer and batch)
    3. activations
    4. dropout
    """

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.input_size])

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size([self.output_size])

    def __init__(
        self,
        input_size: int,
        input_rank: int = 2,
        output_size: int = 256,
        use_bias: bool = True,
        weights_initializer: str = "xavier_uniform",
        bias_initializer: str = "zeros",
        norm: Optional[str] = None,
        norm_params: Optional[Dict] = None,
        activation: str = "relu",
        dropout: float = 0,
    ):
        super().__init__()

        self.layers = ModuleList()
        self.input_size = input_size
        self.output_size = output_size

        fc = Linear(in_features=input_size, out_features=output_size, bias=use_bias)
        self.layers.append(fc)

        weights_initializer = initializer_registry[weights_initializer]
        weights_initializer(fc.weight)

        if use_bias:
            bias_initializer = initializer_registry[bias_initializer]
            bias_initializer(fc.bias)

        if norm is not None:
            norm_params = norm_params or {}
            self.layers.append(create_norm_layer(norm, input_rank, output_size, **norm_params))

        # Dict for activation objects in pytorch?
        self.layers.append(activations[activation]())

        if dropout > 0:
            self.layers.append(Dropout(dropout))

    def forward(self, inputs, mask=None):
        hidden = inputs
        for layer in self.layers:
            hidden = layer(hidden)

        return hidden


class FCStack(LudwigModule):
    """A stack of FCLayers.

    The specification of each FCLayer is specified by the `layers` dictionary parameter, whose keys correspond with an
    FCLayer's constructor arguments, i.e.

    [
        {"input_size": 2, "output_size": 4},
        {"output_size": 4, "use_bias": False},
    ]

    `default_*` parameters dictate default values to use for each FCLayer, if not specified by `layers`. If `layers` is
    `None`, then a stack of size `num_layers` of `FCLayer`s configured with all of the `default_*` parameters is used.

    If `layers` is None and `num_layers` is 0, then there are no fully connected layers and this module serves as a
    trivial passthrough.
    """

    def __init__(
        self,
        first_layer_input_size: int,
        layers: Optional[List[Dict]] = None,
        num_layers: int = 1,
        default_input_rank: int = 2,
        default_output_size: int = 256,
        default_use_bias: bool = True,
        default_weights_initializer: str = "xavier_uniform",
        default_bias_initializer: str = "zeros",
        default_norm: Optional[str] = None,
        default_norm_params: Optional[Dict] = None,
        default_activation: str = "relu",
        default_dropout: float = 0,
        residual: bool = False,
        **kwargs,
    ):
        super().__init__()
        self.input_size = first_layer_input_size

        self.norm_layer = None
        if default_norm is not None:
            norm_params = default_norm_params or {}
            self.norm_layer = create_norm_layer(default_norm, default_input_rank, self.input_size, **norm_params)

        self.dropout = None
        if default_dropout > 0:
            self.dropout = torch.nn.Dropout(default_dropout)

        if layers is None:
            self.layers = []
            for i in range(num_layers):
                self.layers.append({})
        else:
            # deep copy the layer definitions so that we don't modify the original
            self.layers = deepcopy(layers)

        if len(self.layers) > 0 and "input_size" not in self.layers[0]:
            self.layers[0]["input_size"] = first_layer_input_size
        for i, layer in enumerate(self.layers):
            if i != 0:
                layer["input_size"] = self.layers[i - 1]["output_size"]
            if "input_rank" not in layer:
                layer["input_rank"] = default_input_rank
            if "output_size" not in layer:
                layer["output_size"] = default_output_size
            if "use_bias" not in layer:
                layer["use_bias"] = default_use_bias
            if "weights_initializer" not in layer:
                layer["weights_initializer"] = default_weights_initializer
            if "bias_initializer" not in layer:
                layer["bias_initializer"] = default_bias_initializer
            if "norm" not in layer:
                layer["norm"] = default_norm
            if "norm_params" not in layer:
                layer["norm_params"] = default_norm_params
            if "activation" not in layer:
                layer["activation"] = default_activation
            if "dropout" not in layer:
                layer["dropout"] = default_dropout

        self.stack = ModuleList()

        for i, layer in enumerate(self.layers):
            self.stack.append(
                FCLayer(
                    input_size=layer["input_size"],
                    input_rank=layer["input_rank"],
                    output_size=layer["output_size"],
                    use_bias=layer["use_bias"],
                    weights_initializer=layer["weights_initializer"],
                    bias_initializer=layer["bias_initializer"],
                    norm=layer["norm"],
                    norm_params=layer["norm_params"],
                    activation=layer["activation"],
                    dropout=layer["dropout"],
                )
            )
        self.residual = residual

    def forward(self, inputs, mask=None):
        hidden = inputs

        if self.norm_layer is not None:
            hidden = self.norm_layer(hidden)

        if self.dropout is not None:
            hidden = self.dropout(hidden)

        prev_fc_layer_size = self.input_size
        for layer in self.stack:
            out = layer(hidden)
            if self.residual and layer.output_size == prev_fc_layer_size:
                hidden = hidden + out
            else:
                hidden = out
            prev_fc_layer_size = layer.layers[0].out_features
        return hidden

    @property
    def num_layers(self) -> int:
        return len(self.layers)

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.input_size])

    @property
    def output_shape(self) -> torch.Size:
        if len(self.stack) > 0:
            return self.stack[-1].output_shape
        return torch.Size([self.input_size])



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
import warnings
from collections.abc import Iterable
from dataclasses import dataclass
from io import BytesIO
from typing import Callable, List, Optional, Tuple, Union

import numpy as np
import torch
import torchvision.transforms.functional as F
from torchvision.io import decode_image, ImageReadMode
from torchvision.models._api import WeightsEnum

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import CROP_OR_PAD, INTERPOLATE
from ludwig.encoders.base import Encoder
from ludwig.utils.data_utils import get_abs_path
from ludwig.utils.fs_utils import get_bytes_obj_from_path
from ludwig.utils.registry import Registry


@dataclass
class TVModelVariant:
    # Model variant identifier
    variant_id: Union[str, int]

    # TorchVision function to create model class
    create_model_function: Callable

    # Torchvision class for model weights
    model_weights: WeightsEnum


logger = logging.getLogger(__name__)

IMAGE_EXTENSIONS = (".png", ".jpg", ".jpeg", ".tiff", ".bmp", ".gif")


@DeveloperAPI
class ResizeChannels(torch.nn.Module):
    def __init__(self, num_channels: int):
        super().__init__()
        self.num_channels = num_channels

    def forward(self, imgs: torch.Tensor):
        original_imgs_shape = imgs.shape
        if len(original_imgs_shape) == 3:  # if shape is (C, H, W), add batch dimension
            imgs = imgs.unsqueeze(0)

        channels = imgs.shape[1]
        if channels > self.num_channels:
            # take the first `self.num_channels` channels
            imgs = imgs[:, : self.num_channels, :, :]
        elif channels < self.num_channels:
            # repeat and use the first `self.num_channels` channels
            imgs = imgs.repeat(1, (self.num_channels // channels) + 1, 1, 1)[:, : self.num_channels, :, :]

        if len(original_imgs_shape) == 3:  # if shape was (C, H, W), remove batch dimension
            return imgs[0]
        return imgs


@DeveloperAPI
def get_gray_default_image(num_channels: int, height: int, width: int) -> np.ndarray:
    return np.full((num_channels, height, width), 128, dtype=np.float32)


@DeveloperAPI
def get_average_image(image_lst: List[np.ndarray]) -> np.array:
    return np.mean([x for x in image_lst if x is not None], axis=(0), dtype=np.float32)


@DeveloperAPI
def is_image(src_path: str, img_entry: Union[bytes, str], column: str) -> bool:
    if not isinstance(img_entry, str):
        return False
    try:
        import imghdr

        path = get_abs_path(src_path, img_entry)
        bytes_obj = get_bytes_obj_from_path(path)
        if isinstance(bytes_obj, bytes):
            return imghdr.what(None, bytes_obj) is not None
        return imghdr.what(bytes_obj) is not None
    except AttributeError:
        # An AttributeError is raised when an image doesn't exist in the dataset, and we want to silence those errors.
        return False
    except Exception as e:
        logger.warning(f"While assessing potential image in is_image() for column {column}, encountered exception: {e}")
        return False


@DeveloperAPI
def is_image_score(src_path, img_entry, column: str):
    """Used for AutoML For image inference, want to bias towards both readable images, but also account for
    unreadable (i.e. expired) urls with image extensions."""
    if is_image(src_path, img_entry, column):
        return 1
    elif isinstance(img_entry, str) and img_entry.lower().endswith(IMAGE_EXTENSIONS):
        return 0.5
    return 0


@DeveloperAPI
def is_torchvision_encoder(encoder_obj: Encoder) -> bool:
    # TODO(travis): do this through an interface rather than conditional logic
    from ludwig.encoders.image.torchvision import TVBaseEncoder

    return isinstance(encoder_obj, TVBaseEncoder)


@DeveloperAPI
def get_image_read_mode_from_num_channels(num_channels: int) -> ImageReadMode:
    """Returns the torchvision.io.ImageReadMode corresponding to the number of channels.

    If num_channels is not recognized, returns ImageReadMode.UNCHANGED.
    """
    mode = ImageReadMode.UNCHANGED
    if num_channels == 1:
        mode = ImageReadMode.GRAY
    elif num_channels == 2:
        mode = ImageReadMode.GRAY_ALPHA
    elif num_channels == 3:
        mode = ImageReadMode.RGB
    elif num_channels == 4:
        mode = ImageReadMode.RGB_ALPHA
    return mode


@DeveloperAPI
def read_image_from_path(
    path: str, num_channels: Optional[int] = None, return_num_bytes=False
) -> Union[Optional[torch.Tensor], Tuple[Optional[torch.Tensor], int]]:
    """Reads image from path.

    Useful for reading from a small number of paths. For more intensive reads, use backend.read_binary_files instead. If
    `return_num_bytes` is True, returns a tuple of (image, num_bytes).
    """
    bytes_obj = get_bytes_obj_from_path(path)
    image = read_image_from_bytes_obj(bytes_obj, num_channels)
    if return_num_bytes:
        if bytes_obj is not None:
            num_bytes = len(bytes_obj)
        else:
            num_bytes = None
        return image, num_bytes
    else:
        return image


@DeveloperAPI
def read_image_from_bytes_obj(
    bytes_obj: Optional[bytes] = None, num_channels: Optional[int] = None
) -> Optional[torch.Tensor]:
    """Tries to read image as a tensor from the path.

    If the path is not decodable as a PNG, attempts to read as a numpy file. If neither of these work, returns None.
    """
    mode = get_image_read_mode_from_num_channels(num_channels)

    image = read_image_as_png(bytes_obj, mode)
    if image is None:
        image = read_image_as_numpy(bytes_obj)
    if image is None:
        warnings.warn("Unable to read image from bytes object.")
    return image


@DeveloperAPI
def read_image_as_png(
    bytes_obj: Optional[bytes] = None, mode: ImageReadMode = ImageReadMode.UNCHANGED
) -> Optional[torch.Tensor]:
    """Reads image from bytes object from a PNG file."""
    try:
        with BytesIO(bytes_obj) as buffer:
            buffer_view = buffer.getbuffer()
            if len(buffer_view) == 0:
                del buffer_view
                raise Exception("Bytes object is empty. This could be due to a failed load from storage.")
            image = decode_image(torch.frombuffer(buffer_view, dtype=torch.uint8), mode=mode)
            del buffer_view
            return image
    except Exception as e:
        warnings.warn(f"Failed to read image from PNG file. Original exception: {e}")
        return None


@DeveloperAPI
def read_image_as_numpy(bytes_obj: Optional[bytes] = None) -> Optional[torch.Tensor]:
    """Reads image from bytes object from a numpy file."""
    try:
        with BytesIO(bytes_obj) as buffer:
            image = np.load(buffer)
            return torch.from_numpy(image)
    except Exception as e:
        warnings.warn(f"Failed to read image from numpy file. Original exception: {e}")
        return None


@DeveloperAPI
def pad(
    img: torch.Tensor,
    new_size: Union[int, Tuple[int, int]],
) -> torch.Tensor:
    """torchscript-compatible implementation of pad.

    Args:
        img (torch.Tensor): image with shape [..., height, width] to pad
        new_size (Union[int, Tuple[int, int]]): size to pad to. If int, resizes to square image of that size.

    Returns:
        torch.Tensor: padded image of size [..., size[0], size[1]] or [..., size, size] if size is int.
    """
    new_size = to_tuple(new_size)
    old_size = img.shape[-2:]
    pad_size = (torch.tensor(new_size) - torch.tensor(old_size)) / 2
    padding = torch.cat((torch.floor(pad_size), torch.ceil(pad_size)))
    padding[padding < 0] = 0
    padding = [int(x) for x in padding]
    return F.pad(img, padding=padding, padding_mode="edge")


@DeveloperAPI
def crop(
    img: torch.Tensor,
    new_size: Union[int, Tuple[int, int]],
) -> torch.Tensor:
    """torchscript-compatible implementation of crop.

    Args:
        img (torch.Tensor): image with shape [..., height, width] to crop
        size (Union[int, Tuple[int, int]]): size to crop to. If int, crops to square image of that size.

    Returns:
        torch.Tensor: cropped image of size [..., size[0], size[1]] or [..., size, size] if size is int.
    """
    new_size = to_tuple(new_size)
    return F.center_crop(img, output_size=new_size)


@DeveloperAPI
def crop_or_pad(img: torch.Tensor, new_size: Union[int, Tuple[int, int]]):
    """torchscript-compatible implementation of resize using constants.CROP_OR_PAD.

    Args:
        img (torch.Tensor): image with shape [..., height, width] to resize
        new_size (Union[int, Tuple[int, int]]): size to resize to. If int, resizes to square image of that size.

    Returns:
        torch.Tensor: resized image of size [..., size[0], size[1]] or [..., size, size] if size is int.
    """
    new_size = to_tuple(new_size)
    if list(new_size) == list(img.shape[-2:]):
        return img
    img = pad(img, new_size)
    img = crop(img, new_size)
    return img


@DeveloperAPI
def resize_image(
    img: torch.Tensor,
    new_size: Union[int, Tuple[int, int]],
    resize_method: str,
    crop_or_pad_constant: str = CROP_OR_PAD,
    interpolate_constant: str = INTERPOLATE,
) -> torch.Tensor:
    """torchscript-compatible implementation of resize.

    Args:
        img (torch.Tensor): image with shape [..., height, width] to resize
        new_size (Union[int, Tuple[int, int]]): size to resize to. If int, resizes to square image of that size.
        resize_method (str): method to use for resizing. Either constants.CROP_OR_PAD or constants.INTERPOLATE.

    Returns:
        torch.Tensor: resized image of size [..., size[0], size[1]] or [..., size, size] if size is int.
    """
    new_size = to_tuple(new_size)
    if list(img.shape[-2:]) != list(new_size):
        if resize_method == crop_or_pad_constant:
            return crop_or_pad(img, new_size)
        elif resize_method == interpolate_constant:
            return F.resize(img, new_size)
        raise ValueError(f"Invalid image resize method: {resize_method}")
    return img


@DeveloperAPI
def grayscale(img: torch.Tensor) -> torch.Tensor:
    """Grayscales RGB image."""
    return F.rgb_to_grayscale(img)


@DeveloperAPI
def num_channels_in_image(img: torch.Tensor):
    """Returns number of channels in image."""
    if img is None or img.ndim < 2:
        raise ValueError("Invalid image data")

    if img.ndim == 2:
        return 1
    else:
        return img.shape[0]


@DeveloperAPI
def to_tuple(v: Union[int, Tuple[int, int]]) -> Tuple[int, int]:
    """Converts int or tuple to tuple of ints."""
    if torch.jit.isinstance(v, int):
        return v, v
    else:
        return v


@DeveloperAPI
def to_np_tuple(prop: Union[int, Iterable]) -> np.ndarray:
    """Creates a np array of length 2 from a Conv2D property.

    E.g., stride=(2, 3) gets converted into np.array([2, 3]), where the
    height_stride = 2 and width_stride = 3. stride=2 gets converted into
    np.array([2, 2]).
    """
    if type(prop) == int:
        return np.ones(2).astype(int) * prop
    elif isinstance(prop, Iterable) and len(prop) == 2:
        return np.array(list(prop)).astype(int)
    elif type(prop) == np.ndarray and prop.size == 2:
        return prop.astype(int)
    else:
        raise TypeError(f"prop must be int or iterable of length 2, but is {prop}.")


@DeveloperAPI
def get_img_output_shape(
    img_height: int,
    img_width: int,
    kernel_size: Union[int, Tuple[int]],
    stride: Union[int, Tuple[int]],
    padding: Union[int, Tuple[int], str],
    dilation: Union[int, Tuple[int]],
) -> Tuple[int]:
    """Returns the height and width of an image after a 2D img op.

    Currently supported for Conv2D, MaxPool2D and AvgPool2d ops.
    """
    if padding == "same":
        return (img_height, img_width)
    elif padding == "valid":
        padding = np.zeros(2)
    else:
        padding = to_np_tuple(padding)

    kernel_size = to_np_tuple(kernel_size)
    stride = to_np_tuple(stride)
    dilation = to_np_tuple(dilation)
    shape = np.array([img_height, img_width])

    out_shape = np.floor(((shape + 2 * padding - dilation * (kernel_size - 1) - 1) / stride) + 1)

    return tuple(out_shape.astype(int))


torchvision_model_registry = Registry()


def register_torchvision_model_variants(variants: List[TVModelVariant]):
    def wrap(cls):
        # prime with empty placeholder
        torchvision_model_registry[cls.torchvision_model_type] = {}

        # register each variant
        for variant in variants:
            torchvision_model_registry[cls.torchvision_model_type][variant.variant_id] = variant
        return cls

    return wrap



# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Union


@dataclass
class DatasetConfig:
    """The configuration of a Ludwig dataset."""

    # The version of the dataset.
    version: str

    # The name of the dataset. Make this a valid python module name, should not contain spaces or dashes.
    name: str

    # The readable description of the dataset
    description: str = ""

    # Optional. The (suggested) output features for this dataset. Helps users discover new datasets and filter for
    # relevance to a specific machine learning setting.
    output_features: List[dict] = field(default_factory=list)

    # The kaggle competition this dataset belongs to, or None if this dataset is not hosted by a Kaggle competition.
    kaggle_competition: Optional[str] = None

    # The kaggle dataset ID, or None if this dataset if not hosted by Kaggle.
    kaggle_dataset_id: Optional[str] = None

    # The list of URLs to download.
    download_urls: Union[str, List[str]] = field(default_factory=list)

    # The list of file archives which will be downloaded. If download_urls contains a filename with extension, for
    # example https://domain.com/archive.zip, then archive_filenames does not need to be specified.
    archive_filenames: Union[str, List[str]] = field(default_factory=list)

    # The names of files in the dataset (after extraction). Glob-style patterns are supported, see
    # https://docs.python.org/3/library/glob.html
    dataset_filenames: Union[str, List[str]] = field(default_factory=list)

    # If the dataset contains separate files for training, testing, or validation. Glob-style patterns are supported,
    # see https://docs.python.org/3/library/glob.html
    train_filenames: Union[str, List[str]] = field(default_factory=list)
    validation_filenames: Union[str, List[str]] = field(default_factory=list)
    test_filenames: Union[str, List[str]] = field(default_factory=list)

    # If the dataset contains additional referenced files or directories (ex. images or audio) list them here and they
    # will be copied to the same location as the processed dataset. Glob-style patterns are supported,
    # see https://docs.python.org/3/library/glob.html
    preserve_paths: Union[str, List[str]] = field(default_factory=list)

    # Optionally verify integrity of the dataset by providing sha256 checksums for important files. Maps filename to
    # sha256 digest.  Use `sha256sum <filename>` on linux, `shasum -a 256 <filename>` on Mac to get checksums.
    # If verification fails, loading the dataset will fail with a ValueError.
    # If no sha256 digests are in the config, a warning is logged and the dataset will load without verification.
    sha256: Dict[str, str] = field(default_factory=dict)

    # List of column names, for datasets which do not have column names. If specified, will override the column names
    # already present in the dataset.
    columns: List[str] = field(default_factory=list)

    # Optional dictionary which maps column name to column type. Column's will be converted to the requested type, or
    # will be inferred from the dataset by default.
    column_types: Dict[str, str] = field(default_factory=dict)

    # The loader module and class to use, relative to ludwig.datasets.loaders. Only change this if the dataset requires
    # processing which is not handled by the default loader.
    loader: str = "dataset_loader.DatasetLoader"



# Copyright (c) 2020 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import os
from functools import partial
from unittest import mock

import numpy as np
import pandas as pd
import pytest
import torch

from ludwig.api import LudwigModel
from ludwig.constants import BATCH_SIZE, DECODER, NAME, TRAINER
from tests.integration_tests.utils import (
    binary_feature,
    category_feature,
    generate_data,
    RAY_BACKEND_CONFIG,
    set_feature,
    text_feature,
)


def random_binary_logits(*args, num_predict_samples, **kwargs):
    # Produce an even mix of True and False predictions, as the model may be biased
    # towards one direction without training
    return torch.tensor(np.random.uniform(low=-1.0, high=1.0, size=(num_predict_samples,)), dtype=torch.float32)


def random_set_logits(*args, num_predict_samples, vocab_size, pct_positive, **kwargs):
    # Produce a desired mix of predictions based on the pct_positive, as the model may be biased
    # towards one direction without training
    num_positive = int(num_predict_samples * pct_positive)
    num_negative = num_predict_samples - num_positive
    negative_logits = np.random.uniform(low=-1.0, high=-0.1, size=(num_negative, vocab_size))
    positive_logits = np.random.uniform(low=0.1, high=1.0, size=(num_positive, vocab_size))
    logits = np.concatenate([negative_logits, positive_logits], axis=0)
    return torch.tensor(logits, dtype=torch.float32)  # simulate torch model output


@pytest.mark.parametrize(
    "backend",
    [
        pytest.param("local", id="local"),
        pytest.param("ray", id="ray", marks=pytest.mark.distributed),
    ],
)
@pytest.mark.parametrize("distinct_values", [(False, True), ("No", "Yes")])
def test_binary_predictions(tmpdir, backend, distinct_values, ray_cluster_2cpu):
    input_features = [
        category_feature(encoder={"vocab_size": 3}),
    ]

    feature = binary_feature()
    output_features = [
        feature,
    ]

    data_csv_path = generate_data(
        input_features,
        output_features,
        os.path.join(tmpdir, "dataset.csv"),
        num_examples=100,
    )
    data_df = pd.read_csv(data_csv_path)

    # Optionally convert bool values to strings, e.g., {'Yes', 'No'}
    false_value, true_value = distinct_values
    data_df[feature[NAME]] = data_df[feature[NAME]].map(lambda x: true_value if x else false_value)
    data_df.to_csv(data_csv_path, index=False)

    config = {
        "input_features": input_features,
        "output_features": output_features,
        TRAINER: {"epochs": 1, BATCH_SIZE: 128},
    }

    patch_args = (
        "ludwig.features.binary_feature.BinaryOutputFeature.logits",
        partial(random_binary_logits, num_predict_samples=len(data_df)),
    )

    preds_df, _ = predict_with_backend(tmpdir, config, data_csv_path, backend, patch_args=patch_args)
    cols = set(preds_df.columns)
    assert f"{feature[NAME]}_predictions" in cols
    assert f"{feature[NAME]}_probabilities_{str(false_value)}" in cols
    assert f"{feature[NAME]}_probabilities_{str(true_value)}" in cols
    assert f"{feature[NAME]}_probability" in cols

    for pred, prob_0, prob_1, prob in zip(
        preds_df[f"{feature[NAME]}_predictions"],
        preds_df[f"{feature[NAME]}_probabilities_{str(false_value)}"],
        preds_df[f"{feature[NAME]}_probabilities_{str(true_value)}"],
        preds_df[f"{feature[NAME]}_probability"],
    ):
        assert pred == false_value or pred == true_value
        if pred == true_value:
            assert prob_1 == prob
        else:
            assert prob_0 == prob
        assert np.allclose(prob_0, 1 - prob_1)


@pytest.mark.parametrize(
    "backend",
    [
        pytest.param("local", id="local"),
        pytest.param("ray", id="ray", marks=pytest.mark.distributed),
    ],
)
@pytest.mark.parametrize("distinct_values", [(0.0, 1.0), (0, 1)])
def test_binary_predictions_with_number_dtype(tmpdir, backend, distinct_values, ray_cluster_2cpu):
    input_features = [
        category_feature(encoder={"vocab_size": 3}),
    ]

    feature = binary_feature()
    output_features = [
        feature,
    ]

    data_csv_path = generate_data(
        input_features,
        output_features,
        os.path.join(tmpdir, "dataset.csv"),
        num_examples=100,
    )
    data_df = pd.read_csv(data_csv_path)

    # Optionally convert bool values to strings, e.g., {'Yes', 'No'}
    false_value, true_value = distinct_values
    data_df[feature[NAME]] = data_df[feature[NAME]].map(lambda x: true_value if x else false_value)
    data_df.to_csv(data_csv_path, index=False)

    config = {
        "input_features": input_features,
        "output_features": output_features,
        TRAINER: {"epochs": 1, BATCH_SIZE: 128},
    }

    patch_args = (
        "ludwig.features.binary_feature.BinaryOutputFeature.logits",
        partial(random_binary_logits, num_predict_samples=len(data_df)),
    )

    preds_df, _ = predict_with_backend(tmpdir, config, data_csv_path, backend, patch_args=patch_args)
    cols = set(preds_df.columns)
    assert f"{feature[NAME]}_predictions" in cols
    assert f"{feature[NAME]}_probabilities_False" in cols
    assert f"{feature[NAME]}_probabilities_True" in cols
    assert f"{feature[NAME]}_probability" in cols

    for pred, prob_0, prob_1, prob in zip(
        preds_df[f"{feature[NAME]}_predictions"],
        preds_df[f"{feature[NAME]}_probabilities_False"],
        preds_df[f"{feature[NAME]}_probabilities_True"],
        preds_df[f"{feature[NAME]}_probability"],
    ):
        assert isinstance(pred, bool)
        if pred:
            assert prob_1 == prob
        else:
            assert prob_0 == prob
        assert np.allclose(prob_0, 1 - prob_1)


@pytest.mark.parametrize("pct_positive", [1.0, 0.5, 0.0])
def test_set_feature_saving(tmpdir, pct_positive):
    backend = "local"
    input_features = [
        text_feature(encoder={"vocab_size": 3}),
    ]

    feature = set_feature(output_feature=True)
    output_features = [
        feature,
    ]

    data_csv_path = generate_data(
        input_features,
        output_features,
        os.path.join(tmpdir, "dataset.csv"),
        num_examples=100,
    )
    data_df = pd.read_csv(data_csv_path)

    config = {
        "input_features": input_features,
        "output_features": output_features,
        TRAINER: {"epochs": 1, BATCH_SIZE: 128},
    }

    patch_args = (
        "ludwig.features.set_feature.SetOutputFeature.logits",
        partial(
            random_set_logits,
            num_predict_samples=len(data_df),
            vocab_size=feature[DECODER]["vocab_size"] + 1,  # +1 for UNK
            pct_positive=pct_positive,
        ),
    )

    preds_df, ludwig_model = predict_with_backend(tmpdir, config, data_csv_path, backend, patch_args=patch_args)
    cols = set(preds_df.columns)
    assert f"{feature[NAME]}_predictions" in cols
    assert f"{feature[NAME]}_probabilities" in cols

    backend = ludwig_model.backend
    backend.df_engine.to_parquet(preds_df, os.path.join(tmpdir, "preds.parquet"))  # test saving


def predict_with_backend(tmpdir, config, data_csv_path, backend, patch_args=None):
    if backend == "ray":
        backend = RAY_BACKEND_CONFIG
        backend["processor"]["type"] = "dask"

    ludwig_model = LudwigModel(config, backend=backend)
    _, _, output_directory = ludwig_model.train(
        dataset=data_csv_path,
        output_directory=os.path.join(tmpdir, "output"),
    )
    # Check that metadata JSON saves and loads correctly
    ludwig_model = LudwigModel.load(os.path.join(output_directory, "model"))

    if patch_args is not None:
        with mock.patch(*patch_args):
            preds_df, _ = ludwig_model.predict(dataset=data_csv_path)
    else:
        preds_df, _ = ludwig_model.predict(dataset=data_csv_path)

    return preds_df, ludwig_model
